# -*- coding: utf-8 -*-
r"""non-linear least squares fitting
"""
import numbers
import warnings
from collections import namedtuple

import numpy as np

from lmfit import Minimizer, Parameters, minimize

from .plot import PlotFit
from .tools import residual_wrapper


class Fitter(PlotFit):
    u"""Wrapper for LMFIT, which is a high-level extension for
    scipy.optimize.leastsq. Performs Non-Linear Least Squares fitting using
    the Levenberg-Marquardt method.

    Parameters
    ----------
    residuals : func
        The residuals function, see description below.

    derivatives : func, optional
        Derivatives function, to compute the Jacobian of the residuals
        function with derivatives across the rows. If this is None, the
        Jacobian will be estimated.

    data : tuple, optional
        Default: None

    params0 : list, optional
        Default: None

    parinfo : list, optional
        Default: None

    ftol : float, optional
        Default: 1e-10

    xtol : float, optional
        Default: 1e-10

    epsfcn : float, optional
        Default: 2.2204460492503131e-16

    stepfactor : float, optional
        Default: 100.0

    covtol : float, optional
        Default: 1e-14

    maxiter : int, optional
        Default: 200

    maxfev : int, optional
        Default: 0

    nofinitecheck : bool, optional
        Default: False

    nan_policy : str, optional
        Default: 'omit'. Determines how NaN values are handled: 'raise',
        'propagate' or 'omit'.

    Notes
    -----
    Objects of this class are callable, returning the fitted parameters.

    **Residuals function**
    The residuals function must return an ndarray with weighted deviations
    between the model and the data. It takes two arguments, a list of the
    parameter values and a reference for the attribute :attr:`data`, a tuple
    e.g. ``(x, y, err)``. In a typical scientific problem the residuals should
    be weighted so that each deviate has a Gaussian sigma of 1.0.  If ``x``
    represents the independent variable, ``y`` represents an intensity for
    each value of ``x``, and ``err`` represents the error, then the deviates
    could be calculated as follows:

    .. math::

       d = (y - f(x)) / err

    where *f* is the model function. If *err* are 1-sigma uncertainties in
    ``y``, then

    .. math::

       \sum d^2

    is the total chi-squared.  :py:meth:`Fitter.fit` will minimize this value.
    ``x``, ``y`` and ``err`` are passed to the residuals function from
    :attr:`data`.

    Attributes
    ----------
    parinfo
    params0
    data
    ftol
    xtol
    gtol
    epsfcn
    stepfactor
    covtol
    maxiter
    maxfev
    params
    xerror
    covar
    chi2_min
    orignorm
    rchi2_min
    stderr
    npar
    nfree
    npegged
    dof
    resid
    niter
    nfev
    status
    message
    residuals
    nofinitecheck

    Methods
    -------
    fit
    plot
    build_param_table
    __call__
    """

    def __init__(self, residuals, derivatives=None, data=None, params0=None, parinfo=None, ftol=1e-10, xtol=1e-10,
                 gtol=1e-10, epsfcn=None, stepfactor=100.0, covtol=1e-14, maxiter=200, maxfev=None,
                 nofinitecheck=False, nan_policy='omit'):

        self._m = 0
        self.result = namedtuple('result', [])
        self.config = namedtuple('config', [])

        self.residuals = residual_wrapper(residuals)
        if derivatives is not None:
            self.deriv = residual_wrapper(derivatives)
        else:
            self.deriv = None

        self.data = data
        self.params0 = params0
        self.parinfo = parinfo
        self.ftol = ftol
        self.xtol = xtol
        self.gtol = gtol
        self.epsfcn = epsfcn
        self.stepfactor = stepfactor
        self.covtol = covtol
        self.maxiter = maxiter
        self.maxfev = maxfev
        self.nofinitecheck = nofinitecheck
        self.nan_policy = nan_policy

    def __call__(self, params0=None):
        if hasattr(self, 'params'):
            return self.params
        elif self.params0 is not None and params0 is None:
            self.fit(self.params0)
            return self.params
        elif params0 is not None:
            self.fit(params0)
            return self.params
        else:
            raise ValueError('params0 is undefined, no fit can be performed')

    @property
    def parinfo(self):
        r"""A list of dicts with parameter constraints, one dict per
        parameter, or None if not given.

        Each dict can have zero or more items with the following keys and
        values:

         ``'fixed'``: bool
            Parameter to be fixed. Default: False.

         ``'limits'``: list
            Two-element list with upper end lower parameter limits or None,
            which indicates that the parameter is not bounded on this side.
            Default: None.
        """
        return self._parinfo

    @parinfo.setter
    def parinfo(self, value):
        if isinstance(value, (list, tuple)):
            if np.all([isinstance(item, (type(None), dict)) for item in value]):
                self._parinfo = value
            else:
                raise ValueError
        elif value is None:
            self._parinfo = None
        else:
            raise ValueError

    @property
    def params(self):
        r"""The fitted parameters. This attribute has the same type as
        :attr:`params0`.
        """
        return self.result.params

    @property
    def params0(self):
        r"""Required attribute. A NumPy array, a tuple or a list with the
        initial parameters values.
        """
        return self._params0

    @params0.setter
    def params0(self, value):
        if isinstance(value, (list, tuple, np.ndarray)):
            self._params0 = value
        elif value is None:
            self._params0 = None
        else:
            raise ValueError

    @property
    def data(self):
        r"""Required attribute. Python object with information for the
        residuals function and the derivatives function. See above.
        """
        return self._data

    @data.setter
    def data(self, value):
        if isinstance(value, tuple):
            self._data = value
        else:
            ValueError

    @property
    def deriv(self):
        return self._deriv

    @deriv.setter
    def deriv(self, value):
        self._deriv = value

    @property
    def ftol(self):
        r"""Relative :math:`\chi^2` convergence criterium. Default: 1e-10
        """
        return self.config.ftol

    @ftol.setter
    def ftol(self, value):
        if isinstance(value, numbers.Number):
            self.config.ftol = value
        else:
            raise ValueError

    @property
    def xtol(self):
        r"""Relative parameter convergence criterium. Default: 1e-10
        """
        return self.config.xtol

    @xtol.setter
    def xtol(self, value):
        if isinstance(value, numbers.Number):
            self.config.xtol = value
        else:
            raise ValueError

    @property
    def gtol(self):
        r"""Orthogonality convergence criterium. Default: 1e-10
        """
        return self.config.gtol

    @gtol.setter
    def gtol(self, value):
        if isinstance(value, numbers.Number):
            self.config.gtol = value
        else:
            raise ValueError

    @property
    def epsfcn(self):
        r"""Finite derivative step size. Default: 2.2204460492503131e-16
        """
        return self.config.epsfcn

    @epsfcn.setter
    def epsfcn(self, value):
        if value is None:
            value = np.finfo(np.float64).eps
        if isinstance(value, numbers.Number):
            self.config.epsfcn = value
        else:
            raise ValueError

    @property
    def stepfactor(self):
        r"""Initial step bound. Default: 100.0
        """
        return self.config.stepfactor

    @stepfactor.setter
    def stepfactor(self, value):
        if isinstance(value, numbers.Number):
            self.config.stepfactor = value
        else:
            raise ValueError

    @property
    def covtol(self):
        r"""(DEPRECIATED) Range tolerance for covariance calculation.
        Default: 1e-14
        """
        warnings.warn('covtol is Depreciated and has no effect', DeprecationWarning)
        return self.config.covtol

    @covtol.setter
    def covtol(self, value):
        if isinstance(value, numbers.Number):
            self.config.covtol = value
        else:
            raise ValueError

    @property
    def maxiter(self):
        r"""(DEPRECIATED) Maximum number of iterations. Default: 200
        """
        warnings.warn('maxiter is Depreciated and has no effect', DeprecationWarning)
        return self.config.maxiter

    @maxiter.setter
    def maxiter(self, value):
        if isinstance(value, int):
            self.config.maxiter = value
        else:
            raise ValueError

    @property
    def maxfev(self):
        r"""Maximum number of function evaluations. Default: 0
        """
        return self.config.maxfev

    @maxfev.setter
    def maxfev(self, value):
        if isinstance(value, int):
            self.config.maxfev = value
        elif value is None:
            self.config.maxfev = 0
        else:
            raise ValueError

    @property
    def nofinitecheck(self):
        r"""(DEPRECIATED) Does not check for finite values. Default: False
        """
        warnings.warn('nofinitecheck is Depreciated and has no effect', DeprecationWarning)
        return self._nofinitecheck

    @nofinitecheck.setter
    def nofinitecheck(self, value):
        if isinstance(value, bool):
            self.config.nofinitecheck = value
        else:
            raise ValueError

    @property
    def nan_policy(self):
        r"""Determines how NaN's are handled by minimizer. Default: 'omit'
        """
        return self._nan_policy

    @nan_policy.setter
    def nan_policy(self, value):
        if isinstance(value, str):
            self._nan_policy = value
        else:
            raise ValueError

    @property
    def npar(self):
        r"""Number of parameters
        """
        try:
            return len(self.params0)
        except TypeError:
            return None

    @property
    def message(self):
        """Success/error message
        """
        return self.result.message

    @property
    def chi2_min(self):
        """Final :math:`\chi^2`
        """
        return self.result.bestnorm

    @property
    def orignorm(self):
        """Initial :math:`\chi^2`.
        """
        return self.result.orignorm

    @property
    def niter(self):
        """Number of iterations
        """
        return self.result.niter

    @property
    def nfev(self):
        """Number of function evaluations
        """
        return self.result.nfev

    @property
    def status(self):
        """Status code of fit passed from scipy.optimize.leastsq
        """
        return self.result.status

    @property
    def nfree(self):
        """Number of free parameters
        """
        return self.result.nfree

    @property
    def npegged(self):
        """Number of fixed parameters
        """
        return self.result.npegged

    @property
    def covar(self):
        """Parameter covariance matrix
        """
        return self.result.covar

    @property
    def resid(self):
        """Residuals
        """
        return self.result.resid

    @property
    def xerror(self):
        """Parameter uncertainties (:math:`1 \sigma`)
        """
        return self.result.xerror

    @property
    def dof(self):
        """Degrees of freedom
        """
        return self._m - self.nfree

    @property
    def rchi2_min(self):
        """Minimum reduced :math:`\chi^2`.
        """
        return self.result.redchi

    @property
    def stderr(self):
        """Standard errors estimated from
        :math:`\sqrt{diag(covar) * \chi^{2}_{reduced}`
        """
        return np.sqrt(np.diagonal(self.covar) * self.rchi2_min)



    def fit(self, params0):
        r"""Perform a fit with the provided parameters.

        Parameters
        ----------
        params0 : list
            Initial fitting parameters

        """
        self.params0 = params0
        p = Parameters()

        if self.parinfo is None:
            self.parinfo = [None] * len(self.params0)
        else:
            assert (len(self.params0) == len(self.parinfo))

        for i, (p0, parin) in enumerate(zip(self.params0, self.parinfo)):
            p.add(name='p{0}'.format(i), value=p0)

            if parin is not None:
                if 'limits' in parin:
                    p['p{0}'.format(i)].set(min=parin['limits'][0])
                    p['p{0}'.format(i)].set(max=parin['limits'][1])
                if 'fixed' in parin:
                    p['p{0}'.format(i)].set(vary=not parin['fixed'])

        if np.all([not value.vary for value in p.values()]):
            raise Exception('All parameters are fixed!')

        self.lmfit_minimizer = Minimizer(self.residuals, p, nan_policy=self.nan_policy, fcn_args=(self.data,))

        self.result.orignorm = np.sum(self.residuals(params0, self.data) ** 2)

        result = self.lmfit_minimizer.minimize(Dfun=self.deriv, method='leastsq', ftol=self.ftol,
                                               xtol=self.xtol, gtol=self.gtol, maxfev=self.maxfev, epsfcn=self.epsfcn,
                                               factor=self.stepfactor)

        self.result.bestnorm = result.chisqr
        self.result.redchi = result.redchi
        self._m = result.ndata
        self.result.nfree = result.nfree
        self.result.resid = result.residual
        self.result.status = result.ier
        self.result.covar = result.covar
        self.result.xerror = [result.params['p{0}'.format(i)].stderr for i in range(len(result.params))]

        self.result.params = [result.params['p{0}'.format(i)].value for i in range(len(result.params))]

        self.result.message = result.message

        self.lmfit_result = result

        if not result.errorbars or not result.success:
            warnings.warn(self.result.message)

        return result.success



import warnings
from collections import OrderedDict

import numpy as np

from ...data import Data
from ..instrument import load_instrument as load_instr


class Neutronpy(Data):
    r"""Loads neutronpy format ascii or hdf5 data file into a Data object

    """

    def __init__(self):
        super(Neutronpy, self).__init__()

    def load(self, filename, build_hkl=True, load_instrument=False):
        r"""Loads data from ascii or hdf5 format file.

        Parameters
        ----------
        filename : str
            Path to file to load

        build_hkl : bool, optional
            Option to build Q = [h, k, l, e, temp]

        load_instrument : bool, optional
            Option to build Instrument from file header

        """
        if filename[-2:].lower() == 'h5':
            self.load_hdf5(filename, build_hkl, load_instrument)
        else:
            with open(filename, 'r') as f:
                first_line = f.readline()
                if 'NeutronPy' in first_line:
                    self.load_ascii(filename, build_hkl, load_instrument)
                else:
                    raise IOError

    def load_hdf5(self, filename, build_hkl=True, load_instrument=False):
        r"""Loads data from HDF5 format file

        Parameters
        ----------
        filename : str
            Path to file to load

        build_hkl : bool, optional
            Option to build Q = [h, k, l, e, temp]

        load_instrument : bool, optional
            Option to build Instrument from file header

        """
        import h5py

        data = OrderedDict()

        with h5py.File(filename, 'r') as f:
            data_root = f['data']

            for key, value in data_root.items():
                data[key] = np.copy(value)

            self._data = data
            self.data_keys = dict((key, str(value)) for key, value in data_root['data_keys'].attrs.items())

            if build_hkl:
                try:
                    self.Q_keys = dict((key, str(value)) for key, value in data_root['Q_keys'].attrs.items())
                except KeyError:
                    warnings.warn('Q_keys could not be built automatically.')

        if load_instrument:
            self.instrument = load_instr(filename, filetype='hdf5')

    def load_ascii(self, filename, build_hkl=True, load_instrument=False):
        r"""Loads data from ascii format file

        Parameters
        ----------
        filename : str
            Path to file to load

        build_hkl : bool, optional
            Option to build Q = [h, k, l, e, temp]

        load_instrument : bool, optional
            Option to build Instrument from file header

        """
        file_header = []
        with open(filename) as f:
            for line in f:
                if '#' in line and 'npy_col_headers' not in line:
                    file_header.append(line.replace('\n', '').replace('# ', ''))
                if 'npy_col_headers' in line:
                    args = next(f).split()
                    col_headers = [head for head in args[1:]]

        args = np.genfromtxt(filename, unpack=True, comments='#', dtype=np.float64)

        data = OrderedDict()
        for head, col in zip(col_headers, args):
            data[head] = col

        self._data = data

        for n, line in enumerate(file_header):
            if 'original_header' in line:
                start = n + 1

            if build_hkl:
                if 'Q_keys' in line:
                    self.Q_keys = eval(line.replace('Q_keys =', ''))

            if 'data_keys' in line:
                self.data_keys = eval(line.replace('data_keys = ', ''))

            if 'def_x' in line:
                self.plot_default_x = line.split('=')[-1].strip()

            if 'def_y' in line:
                self.plot_default_y = line.split('=')[-1].strip()

        if build_hkl and not hasattr(self, 'Q_keys'):
            warnings.warn('Q_keys could not be built automatically.')

        self.file_header = file_header[start:-3]

        if load_instrument:
            try:
                self.instrument = load_instr(filename.split('.')[0] + '.instr', filetype='ascii')
            except IOError:
                warnings.warn('Instrument could not be loaded automatically.')



# -*- coding: utf-8 -*-
r"""Useful constants for neutron scattering calculations, including:

* ``magnetic_form_factors()`` : Magnetic Ion j-values
* ``periodic_table()`` : Periodic table values
* ``scattering_lengths()`` : Neutron scattering lengths
* ``symmetry()`` : Space group information
* ``JOULES_TO_MEV`` : Joules-to-meV conversion factor
* ``BOLTZMANN_IN_MEV_K`` : Boltzmann constant in meV/K
* ``N_A`` : Avogadro constant
* ``neutron_mass`` : Mass of a neutron in grams
* ``e`` : Electric charge of an electron in Coulombs

"""
import json
import os


def magnetic_ion_j():
    r"""Loads j values for Magnetic ions.

    Parameters
    ----------
    None

    Returns
    -------
    magnetic_ion_j : dict
        Database of j-values for magnetic ions

    """
    with open(os.path.join(os.path.dirname(__file__),
                           "database/magnetic_form_factors.json"), 'r') as infile:
        return json.load(infile)


def periodic_table():
    r"""Loads periodic table database.
    mass, and long-form name.

    Parameters
    ----------
    None

    Returns
    -------
    periodic_table : dict
        Database of mass, atomic number, density, mass, and name for all
        elements in the Periodic table

    """
    with open(os.path.join(os.path.dirname(__file__),
                           "database/periodic_table.json"), 'r') as infile:
        return json.load(infile)


def scattering_lengths():
    r"""Loads neutron scattering lengths.

    Parameters
    ----------
    None

    Returns
    -------
    scattering_lengths : dict
        Database of elements containing the absolute, coherent, incoheret, and
        scattering cross-sections and scattering lengths

    """
    with open(os.path.join(os.path.dirname(__file__),
                           "database/scattering_lengths.json"), 'r') as infile:
        return json.load(infile)


def symmetry():
    r"""Loads crystal lattice space groups.

    Parameters
    ----------
    None

    Returns
    -------
    lattice_space_groups : dict
        Database of 230 crystal lattice space groups and their generators

    """
    with open(os.path.join(os.path.dirname(__file__),
                           "database/symmetry.json"), 'r') as infile:
        return json.load(infile)


JOULES_TO_MEV = 1. / 1.6021766208e-19 * 1.e3  # Joules to meV
BOLTZMANN_IN_MEV_K = 8.6173303e-05 * 1.e3  # Boltzmann constant in meV/K
N_A = 6.022140857e+23
neutron_mass = 1.674927211e-24  # mass of a neutron in grams
hbar = 1.054571628e-34  # hbar in m2 kg / s
e = 1.602176487e-19  # coulombs



# -*- coding: utf-8 -*-
import numpy as np

from ..crystal import Sample
from ..energy import Energy
from ..instrument import Instrument
from ..instrument.tas_instrument import TripleAxisInstrument
from ..instrument.tof_instrument import TimeOfFlightInstrument

taz_keys = {'algo': 'method',
            'ana_d': 'ana.d',
            'ana_effic': '',
            'ana_mosaic': 'ana.mosaic',
            'ana_scatter_sense': 'ana.direct',
            'h_coll_after_sample': 'hcol[-2]',
            'h_coll_ana': 'hcol[-1]',
            'h_coll_before_sample': 'hcol[-3]',
            'h_coll_mono': 'hcol[-4]',
            'mono_d': 'mono.d',
            'mono_mosaic': 'mono.mosaic',
            'mono_refl': '',
            'mono_scatter_sense': 'mono.direct',
            'pop_ana_curvh': 'ana.rh',
            'pop_ana_curvv': 'ana.rv',
            'pop_ana_h': 'ana.height',
            'pop_ana_thick': 'ana.depth',
            'pop_ana_use_curvh': '',
            'pop_ana_use_curvv': '',
            'pop_ana_w': 'ana.width',
            'pop_det_h': 'detector.height',
            'pop_det_rect': '',
            'pop_det_w': 'detector.width',
            'pop_dist_ana_det': 'arms[-2]',
            'pop_dist_mono_sample': 'arms[-4]',
            'pop_dist_sample_ana': 'arms[-3]',
            'pop_dist_src_mono': 'arms[-5]',
            'pop_guide_divh': 'arms[-1]',
            'pop_guide_divv': '',
            'pop_mono_curvh': 'mono.rh',
            'pop_mono_curvv': 'mono.rv',
            'pop_mono_h': 'mono.height',
            'pop_mono_thick': 'mono.depth',
            'pop_mono_use_curvh': '',
            'pop_mono_use_curvv': '',
            'pop_mono_w': 'mono.width',
            'pop_sample_cuboid': 'sample.depth',
            'pop_sample_h': 'sample.height',
            'pop_sample_wperpq': '',
            'pop_sample_wq': 'sample.width',
            'pop_src_h': 'guide.height',
            'pop_src_rect': '',
            'pop_src_w': 'guide.width',
            'sample_mosaic': 'sample.mosaic',
            'sample_scatter_sense': 'sample.direct',
            'use_guide': '',
            'v_coll_after_sample': 'vcol[-2]',
            'v_coll_ana': 'vcol[-1]',
            'v_coll_before_sample': 'vcol[-3]',
            'v_coll_mono': 'vcol[-4]'}


def load_instrument(filename, filetype='ascii'):
    r"""Creates Instrument class using input par and cfg files.

    Parameters
    ----------
    filename : str or tuple
        Path to the instrument file. If filetype is 'parcfg' then two
        files should be provided in order of '.par', '.cfg' as a tuple.

    filetype : str, optional
        Default: 'ascii'. Instrument file formats 'ascii', 'parcfg', 'hdf5',
        or 'taz' (xml) are available. 'ascii', 'hdf5' and 'taz' in format
        generated by :py:func:`.load_instrument`. 'taz' format compatible
        with Takin software. See Notes for information about 'parcfg' format
        details.

    Returns
    -------
    setup : obj
        Returns Instrument class object based on the information in the input
        files.

    Notes
    -----
    The format of the ``parfile`` consists of two tab-separated columns, the first
    column containing the values and the second column containing the value
    names preceded by a '%' character:

    +-------+---------+---------------------------------------------------------------------------------+
    | Type  | Name    | Description                                                                     |
    +=======+=========+=================================================================================+
    | float | %DM     | Monochromater d-spacing (Ang^-1)                                                |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %DA     | Analyzer d-spacing (Ang^-1)                                                     |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %ETAM   | Monochromator mosaic (arc min)                                                  |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %ETAA   | Analyzer mosaic (arc min)                                                       |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %ETAS   | Sample mosaic (arc min)                                                         |
    +-------+---------+---------------------------------------------------------------------------------+
    | int   | %SM     | Scattering direction of monochromator (+1 clockwise, -1 counterclockwise)       |
    +-------+---------+---------------------------------------------------------------------------------+
    | int   | %SS     | Scattering direction of sample (+1 clockwise, -1 counterclockwise)              |
    +-------+---------+---------------------------------------------------------------------------------+
    | int   | %SA     | Scattering direction of analyzer (+1 clockwise, -1 counterclockwise)            |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %K      | Fixed wavevector (incident or final) of neutrons                                |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %ALPHA1 | Horizontal collimation of in-pile collimator (arc min)                          |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %ALPHA2 | Horizontal collimation of collimator between monochromator and sample (arc min) |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %ALPHA3 | Horizontal collimation of collimator between sample and analyzer (arc min)      |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %ALPHA4 | Horizontal collimation of collimator between analyzer and detector (arc min)    |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %BETA1  | Vertical collimation of in-pile collimator (arc min)                            |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %BETA2  | Vertical collimation of collimator between monochromator and sample (arc min)   |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %BETA3  | Vertical collimation of collimator between sample and analyzer (arc min)        |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %BETA4  | Vertical collimation of collimator between analyzer and detector (arc min)      |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %AS     | Sample lattice constant a (Ang)                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %BS     | Sample lattice constant b (Ang)                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %CS     | Sample lattice constant c (Ang)                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %AA     | Sample lattice angle alpha (deg)                                                |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %BB     | Sample lattice angle beta (deg)                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %CC     | Sample lattice angle gamma (deg)                                                |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %AX     | Sample orientation vector u_x (r.l.u.)                                          |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %AY     | Sample orientation vector u_y (r.l.u.)                                          |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %AZ     | Sample orientation vector u_z (r.l.u.)                                          |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %BX     | Sample orientation vector v_x (r.l.u.)                                          |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %BY     | Sample orientation vector v_y (r.l.u.)                                          |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %BZ     | Sample orientation vector v_z (r.l.u.)                                          |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %QX     |                                                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %QY     |                                                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %QZ     |                                                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %EN     |                                                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %dqx    |                                                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %dqy    |                                                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %dqz    |                                                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %de     |                                                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %gh     |                                                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %gk     |                                                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %gl     |                                                                                 |
    +-------+---------+---------------------------------------------------------------------------------+
    | float | %gmod   |                                                                                 |
    +-------+---------+---------------------------------------------------------------------------------+

    The format of the ``cfgfile`` (containing values necessary for Popovici type
    calculations) can consists of a single column of values, or two
    tab-separated columns, the first column containing the values and the
    second column containing the value descriptions preceded by a '%' character.
    The values MUST be in the following order:

    +-------+-------------------------------------------------------+
    | Type  | Description                                           |
    +=======+=======================================================+
    | float | =0 for circular source, =1 for rectangular source     |
    +-------+-------------------------------------------------------+
    | float | width/diameter of the source (cm)                     |
    +-------+-------------------------------------------------------+
    | float | height/diameter of the source (cm)                    |
    +-------+-------------------------------------------------------+
    | float | =0 No Guide, =1 for Guide                             |
    +-------+-------------------------------------------------------+
    | float | horizontal guide divergence (minutes/Angs)            |
    +-------+-------------------------------------------------------+
    | float | vertical guide divergence (minutes/Angs)              |
    +-------+-------------------------------------------------------+
    | float | =0 for cylindrical sample, =1 for cuboid sample       |
    +-------+-------------------------------------------------------+
    | float | sample width/diameter perp. to Q (cm)                 |
    +-------+-------------------------------------------------------+
    | float | sample width/diameter along Q (cm)                    |
    +-------+-------------------------------------------------------+
    | float | sample height (cm)                                    |
    +-------+-------------------------------------------------------+
    | float | =0 for circular detector, =1 for rectangular detector |
    +-------+-------------------------------------------------------+
    | float | width/diameter of the detector (cm)                   |
    +-------+-------------------------------------------------------+
    | float | height/diameter of the detector (cm)                  |
    +-------+-------------------------------------------------------+
    | float | thickness of monochromator (cm)                       |
    +-------+-------------------------------------------------------+
    | float | width of monochromator (cm)                           |
    +-------+-------------------------------------------------------+
    | float | height of monochromator (cm)                          |
    +-------+-------------------------------------------------------+
    | float | thickness of analyser (cm)                            |
    +-------+-------------------------------------------------------+
    | float | width of analyser (cm)                                |
    +-------+-------------------------------------------------------+
    | float | height of analyser (cm)                               |
    +-------+-------------------------------------------------------+
    | float | distance between source and monochromator (cm)        |
    +-------+-------------------------------------------------------+
    | float | distance between monochromator and sample (cm)        |
    +-------+-------------------------------------------------------+
    | float | distance between sample and analyser (cm)             |
    +-------+-------------------------------------------------------+
    | float | distance between analyser and detector (cm)           |
    +-------+-------------------------------------------------------+
    | float | horizontal curvature of monochromator 1/radius (cm-1) |
    +-------+-------------------------------------------------------+
    | float | vertical curvature of monochromator (cm-1) was 0.013  |
    +-------+-------------------------------------------------------+
    | float | horizontal curvature of analyser (cm-1) was 0.078     |
    +-------+-------------------------------------------------------+
    | float | vertical curvature of analyser (cm-1)                 |
    +-------+-------------------------------------------------------+
    | float | distance monochromator-monitor                        |
    +-------+-------------------------------------------------------+
    | float | width monitor (cm)                                    |
    +-------+-------------------------------------------------------+
    | float | height monitor (cm)                                   |
    +-------+-------------------------------------------------------+

    """
    if filetype == 'parcfg':
        parfile, cfgfile = filename
        with open(parfile, "r") as f:
            lines = f.readlines()
            par = {}
            for line in lines:
                rows = line.split()
                par[rows[1][1:].lower()] = float(rows[0])

        with open(cfgfile, "r") as f:
            lines = f.readlines()
            cfg = []
            for line in lines:
                rows = line.split()
                cfg.append(float(rows[0]))

        if par['sm'] == par['ss']:
            dir1 = -1
        else:
            dir1 = 1

        if par['ss'] == par['sa']:
            dir2 = -1
        else:
            dir2 = 1

        if par['kfix'] == 2:
            infin = -1
        else:
            infin = par['kfix']

        hcol = [par['alpha1'], par['alpha2'], par['alpha3'], par['alpha4']]
        vcol = [par['beta1'], par['beta2'], par['beta3'], par['beta4']]

        nsou = cfg[0]  # =0 for circular source, =1 for rectangular source.
        if nsou == 0:
            ysrc = cfg[1] / 4  # width/diameter of the source [cm].
            zsrc = cfg[2] / 4  # height/diameter of the source [cm].
        else:
            ysrc = cfg[1] / np.sqrt(12)  # width/diameter of the source [cm].
            zsrc = cfg[2] / np.sqrt(12)  # height/diameter of the source [cm].

        flag_guide = cfg[3]  # =0 for no guide, =1 for guide.
        guide_h = cfg[4]  # horizontal guide divergence [mins/Angs]
        guide_v = cfg[5]  # vertical guide divergence [mins/Angs]
        if flag_guide == 1:
            alpha_guide = np.pi / 60. / 180. * 2 * np.pi * guide_h / par['k']
            alpha0 = hcol[0] * np.pi / 60. / 180.
            if alpha_guide <= alpha0:
                hcol[0] = 2. * np.pi / par['k'] * guide_h
            beta_guide = np.pi / 60. / 180. * 2 * np.pi * guide_v / par['k']
            beta0 = vcol[0] * np.pi / 60. / 180.
            if beta_guide <= beta0:
                vcol[0] = 2. * np.pi / par['k'] * guide_v

        nsam = cfg[6]  # =0 for cylindrical sample, =1 for cuboid sample.
        if nsam == 0:
            xsam = cfg[7] / 4  # sample width/diameter perp. to Q [cm].
            ysam = cfg[8] / 4  # sample width/diameter along Q [cm].
            zsam = cfg[9] / 4  # sample height [cm].
        else:
            xsam = cfg[7] / np.sqrt(12)  # sample width/diameter perp. to Q [cm].
            ysam = cfg[8] / np.sqrt(12)  # sample width/diameter along Q [cm].
            zsam = cfg[9] / np.sqrt(12)  # sample height [cm].

        ndet = cfg[10]  # =0 for circular detector, =1 for rectangular detector.
        if ndet == 0:
            ydet = cfg[11] / 4  # width/diameter of the detector [cm].
            zdet = cfg[12] / 4  # height/diameter of the detector [cm].
        else:
            ydet = cfg[11] / np.sqrt(12)  # width/diameter of the detector [cm].
            zdet = cfg[12] / np.sqrt(12)  # height/diameter of the detector [cm].

        xmon = cfg[13]  # thickness of monochromator [cm].
        ymon = cfg[14]  # width of monochromator [cm].
        zmon = cfg[15]  # height of monochromator [cm].

        xana = cfg[16]  # thickness of analyser [cm].
        yana = cfg[17]  # width of analyser [cm].
        zana = cfg[18]  # height of analyser [cm].

        L0 = cfg[19]  # distance between source and monochromator [cm].
        L1 = cfg[20]  # distance between monochromator and sample [cm].
        L2 = cfg[21]  # distance between sample and analyser [cm].
        L3 = cfg[22]  # distance between analyser and detector [cm].

        romh = par['sm'] * cfg[23]  # horizontal curvature of monochromator 1/radius [cm-1].
        romv = par['sm'] * cfg[24]  # vertical curvature of monochromator [cm-1].
        roah = par['sa'] * cfg[25]  # horizontal curvature of analyser [cm-1].
        roav = par['sa'] * cfg[26]  # vertical curvature of analyser [cm-1].
        inv_rads = [romh, romv, roah, roav]
        for n, inv_rad in enumerate(inv_rads):
            if inv_rad == 0:
                inv_rads[n] = 1.e6
            else:
                inv_rads[n] = 1. / inv_rad
        [romh, romv, roah, roav] = inv_rads

        L1mon = cfg[27]  # distance monochromator monitor [cm]
        monitorw = cfg[28] / np.sqrt(12)  # monitor width [cm]
        monitorh = cfg[29] / np.sqrt(12)  # monitor height [cm]

        # -------------------------------------------------------------------------

        energy = Energy(wavevector=par['k'])

        sample = Sample(par['as'], par['bs'], par['cs'],
                        par['aa'], par['bb'], par['cc'],
                        par['etas'])
        sample.u = [par['ax'], par['ay'], par['az']]
        sample.v = [par['bx'], par['by'], par['bz']]
        sample.shape = np.diag([xsam, ysam, zsam])

        setup = Instrument(energy.energy, sample, hcol, vcol,
                           2 * np.pi / par['dm'], par['etam'],
                           2 * np.pi / par['da'], par['etaa'])

        setup.method = 1
        setup.dir1 = dir1
        setup.dir2 = dir2
        setup.mondir = par['sm']
        setup.infin = infin
        setup.arms = [L0, L1, L2, L3, L1mon]
        setup.guide.width = ysrc
        setup.guide.height = zsrc

        setup.detector.width = ydet
        setup.detector.height = zdet

        setup.mono.depth = xmon
        setup.mono.width = ymon
        setup.mono.height = zmon
        setup.mono.rv = romv
        setup.mono.rh = romh

        setup.ana.depth = xana
        setup.ana.width = yana
        setup.ana.height = zana
        setup.ana.rv = roav
        setup.ana.rh = roah

        setup.monitor.width = monitorw
        setup.monitor.height = monitorh

        return setup

    elif filetype == 'ascii':
        with open(filename, 'r') as f:
            lines = []
            for line in f:
                value = line.replace('\n', '').replace('instrument.', '').split('=')
                left = value[0].split('.')
                for n, el in enumerate(left):
                    left[n] = el.strip()

                right = value[1].strip()
                lines.append([left, right])

        setup = Instrument()

        for line in lines:
            attr = line[0]
            value = line[1]
            if '[' in value:
                value = value.replace(']', '').replace('[', '').split(' ')
                value = [float(item) for item in value]
            else:
                try:
                    value = float(line[1])
                except ValueError:
                    pass

            if len(attr) == 1:
                setattr(setup, attr[0], value)
            else:
                subobj = getattr(setup, attr[0])
                setattr(subobj, attr[1], value)
                setattr(setup, attr[0], subobj)

        return setup

    elif filetype == 'hdf5':
        import h5py

        setup = Instrument()

        with h5py.File(filename, mode='r') as f:
            instrument = f['instrument']
            groups = list(instrument.keys())

            for key, value in instrument.attrs.items():
                setattr(setup, key, value)

            for grp in groups:
                subobj = getattr(setup, grp)
                for key, value in instrument[grp].attrs.items():
                    if isinstance(value, np.bytes_):
                        value = value.decode('UTF-8')
                    setattr(subobj, key, value)
                setattr(setup, grp, subobj)

        return setup

    elif filetype == 'taz':
        import xml.etree.ElementTree as et

        tree = et.ElementTree(file=filename)
        taz = tree.getroot()
        reso = taz.find('reso')

        setup = Instrument()

        values = dict()

        for key, value in taz_keys.items():
            values[value] = reso.find(key).text

        hcol = [float(values[key]) for key in ['hcol[-4]', 'hcol[-3]', 'hcol[-2]', 'hcol[-1]'] if len(values[key]) > 0]
        vcol = [float(values[key]) for key in ['vcol[-4]', 'vcol[-3]', 'vcol[-2]', 'vcol[-1]'] if len(values[key]) > 0]
        arms = [float(values[key]) for key in ['arms[-5]', 'arms[-4]', 'arms[-3]', 'arms[-2]', 'arms[-1]'] if
                len(values[key]) > 0]

        for key, value in zip(['hcol', 'vcol', 'arms'], [hcol, vcol, arms]):
            setattr(setup, key, value)

        for key, value in values.items():
            if '[' not in key:
                if '.' in key:
                    subobj = getattr(setup, key.split('.')[0])
                    setattr(subobj, key.split('.')[1], value)
                    setattr(setup, key.split('.')[0], subobj)
                else:
                    setattr(setup, key, value)

        return setup

    else:
        raise ValueError("Format not supported. Please use 'ascii', 'hdf5', or 'taz'")


def save_instrument(obj, filename, filetype='ascii', overwrite=False):
    r"""Saves a TAS or TOF instrument configuration into files for loading
    with :py:func:`.load_instrument`.

    Parameters
    ----------
    obj : object
        Instrument object

    filename : str
        Path to file (extension determined by filetype parameter).

    filetype : str, optional
        Default: `'ascii'`. Support for `'ascii'`, `'hdf5'`, or `'taz'`.

    overwrite : bool, optional
        Default: False. If True, overwrites the file, otherwise appends or
        creates new files.

    """
    if isinstance(obj, TripleAxisInstrument):
        save_tas_instrument(obj, filename, filetype, overwrite)
    elif isinstance(obj, TimeOfFlightInstrument):
        save_tof_instrument(obj, filename, filetype, overwrite)


def save_tas_instrument(obj, filename, filetype='ascii', overwrite=False):
    r"""Saves a TAS instrument configuration into files for loading with
    :py:func:`.load_instrument`.

    Parameters
    ----------
    obj : object
        Instrument object

    filename : str
        Path to file (extension determined by filetype parameter).

    filetype : str, optional
        Default: `'ascii'`. Support for `'ascii'`, `'hdf5'`, or `'taz'`.

    overwrite : bool, optional
        Default: False. If True, overwrites the file, otherwise appends or
        creates new files.

    """
    instr_attrs = ['efixed', 'arms', 'hcol', 'vcol', 'method', 'moncor', 'infin']
    mono_attrs = ['tau', 'height', 'width', 'depth', 'direct', 'mosaic', 'vmosaic', 'rh', 'rv']
    ana_attrs = ['tau', 'height', 'width', 'depth', 'direct', 'mosaic', 'vmosaic', 'rh', 'rv', 'horifoc', 'thickness',
                 'Q']
    det_attrs = ['height', 'width', 'depth']
    guide_attrs = ['height', 'width']
    sample_attrs = ['a', 'b', 'c', 'alpha', 'beta', 'gamma', 'u', 'v', 'mosaic', 'vmosaic', 'height', 'width', 'depth',
                    'direct']
    Smooth_attrs = ['X', 'Y', 'Z', 'E']

    if filetype == 'ascii':
        if overwrite:
            mode = 'w+'
        else:
            mode = 'r+'

        lines = []
        for grp_name, attrs in zip(['', 'mono', 'ana', 'detector', 'guide', 'sample', 'Smooth'],
                                   [instr_attrs, mono_attrs, ana_attrs, det_attrs, guide_attrs, sample_attrs,
                                    Smooth_attrs]):
            for attr in attrs:
                value = ''
                if len(grp_name) == 0:
                    try:
                        value = getattr(obj, attr)
                        value = 'instrument.' + str(attr) + ' = ' + str(value) + '\n'
                    except AttributeError:
                        pass
                else:
                    try:
                        value = getattr(getattr(obj, grp_name), attr)
                        value = 'instrument.' + grp_name + '.' + str(attr) + ' = ' + str(value) + '\n'
                    except AttributeError:
                        pass
                if value:
                    lines.append(value)

        with open(filename + '.instr', mode) as f:
            f.writelines(lines)

    elif filetype == 'hdf5':
        import h5py

        if overwrite:
            mode = 'w'
        else:
            mode = 'a'

        with h5py.File(filename + '.h5', mode) as f:
            instrument = f.create_group('instrument')
            mono = instrument.create_group('mono')
            ana = instrument.create_group('ana')
            detector = instrument.create_group('detector')
            guide = instrument.create_group('guide')
            sample = instrument.create_group('sample')
            Smooth = instrument.create_group('Smooth')

            for grp, grp_name, attrs in zip([instrument, mono, ana, detector, guide, sample, Smooth],
                                            ['', 'mono', 'ana', 'detector', 'guide', 'sample', 'Smooth'],
                                            [instr_attrs, mono_attrs, ana_attrs, det_attrs, guide_attrs, sample_attrs,
                                             Smooth_attrs]):
                for attr in attrs:
                    try:
                        if len(grp_name) == 0:
                            value = getattr(obj, attr)
                            if isinstance(value, str):
                                value = value.encode('utf8')
                            grp.attrs.create(attr, value)
                        else:
                            value = getattr(getattr(obj, grp_name), attr)
                            if isinstance(value, str):
                                value = value.encode('utf8')
                            grp.attrs.create(attr, value)
                    except AttributeError:
                        pass

                if len(list(grp.attrs.keys())) == 0:
                    del instrument[grp_name]

    elif filetype == 'taz':
        import xml.etree.ElementTree as et
        from xml.dom import minidom

        def prettify(elem):
            """Return a pretty-printed XML string for the Element.
            """
            rough_string = et.tostring(elem, encoding='utf-8')
            reparsed = minidom.parseString(rough_string)
            return reparsed.toprettyxml(indent="  ")

        if overwrite:
            mode = 'w+'
        else:
            mode = 'r+'

        taz = et.Element('taz')
        reso = et.SubElement(taz, 'reso')

        defaults = [1, 0, 0, 1, 0, 3.355, 45, 1, 5, 3.355, 45, 1, 30, 1e4, 30, 1e4, 1e4, 1e4, 1e4, 1e4, 12, 6, 1,
                    0, 200, 8, .15, 0, 1, 12, 1.5, 0, 3, 1.5, 0, 0, 8, 0.3, 0, 0, 12, 2.5, 5, 1, 10, 200, 115, 85, 15,
                    15]

        for (ele, attr), dflt in zip(taz_keys.items(), defaults):
            subel = et.SubElement(reso, ele)
            value = str(dflt)
            if '.' not in attr:
                if '[' not in attr:
                    try:
                        value = str(getattr(obj, attr))
                    except AttributeError:
                        pass
                else:
                    ind = int(attr[-3:-1])
                    attr = attr[:-4]
                    try:
                        value = str(getattr(obj, attr)[ind])
                    except AttributeError:
                        pass
            else:
                prnt, chld = attr.split('.')
                try:
                    value = str(getattr(getattr(obj, prnt), chld))
                except AttributeError:
                    pass

            subel.text = value

        taz_pretty = prettify(taz)

        with open(filename + '.taz', mode) as f:
            f.write(taz_pretty)

    else:
        raise ValueError("""Format not supported. Please use 'ascii', 'hdf5', or 'taz'""")


def save_tof_instrument(obj, filename, filetype='ascii', overwrite=True):
    r"""Saves a TOF instrument configuration into files for loading with
    :py:func:`.load_instrument`.

    Parameters
    ----------
    obj : object
        Instrument object

    filename : str
        Path to file (extension determined by filetype parameter).

    filetype : str, optional
        Default: `'ascii'`. Support for `'ascii'`, `'hdf5'`, or `'taz'`.

    overwrite : bool, optional
        Default: False. If True, overwrites the file, otherwise appends or
        creates new files.

    """
    instr_attrs = []

    if filetype == 'ascii':
        pass

    elif filetype == 'hdf5':
        pass

    elif filetype == 'taz':

        keys = {'viol_dist_pulse_mono': ['', 1000],
                'viol_dist_mono_sample': ['', 100],
                'viol_dist_sample_det': ['', 500],
                'viol_dist_pulse_mono_sig': ['', 1],
                'viol_dist_mono_sample_sig': ['', 1],
                'viol_dist_sample_det_sig': ['', 1],
                'viol_angle_ph_f': ['', 0],
                'viol_angle_ph_f_sig': ['', 0.4],
                'viol_angle_ph_i': ['', 0],
                'viol_angle_ph_i_sig': ['', 0.4],
                'viol_angle_tt_f_sig': ['', 0.4],
                'viol_angle_tt_i': ['', 0],
                'viol_angle_tt_i_sig': ['', 0.4],
                'viol_time_det_sig': ['', 5],
                'viol_time_mono_sig': ['', 5],
                'viol_time_pulse_sig': ['', 50],
                'viol_det_sph': ['', 1]}

    else:
        raise ValueError("""Format not supported. Please use 'ascii', 'hdf5', or 'taz'""")



# -*- coding: utf-8 -*-
r"""Polarization Tools

Get correction of polarized cross sections
Get He3 decay time

"""


def polarization_correction(data_pp, data_pm, data_mp, data_mm):
    r"""
    """
    pass



from collections import OrderedDict

import numpy as np

from ...data import Data
from ...instrument import Instrument


class Ice(Data):
    r"""Loads ICE (NCNR) format ascii data file.

    """

    def __init__(self):
        super(Ice, self).__init__()

    def load(self, filename, build_hkl=True, load_instrument=False):
        r"""Loads the ICE (NCNR) format ascii data file.

        Parameters
        ----------
        filename : str
            Path to file to open

        build_hkl : bool, optional
            Option to build Q = [h, k, l, e, temp]

        load_instrument : bool, optional
            Option to build Instrument from file header

        """
        with open(filename) as f:
            file_header = []
            for line in f:
                if 'Columns' in line:
                    args = line.split()
                    col_headers = [head for head in args[1:]]
                    break

        args = np.genfromtxt(filename, usecols=(0, 1, 2, 3, 4, 5, 6, 7, 8),
                             unpack=True, comments="#", dtype=np.float64)

        data = OrderedDict()
        for head, col in zip(col_headers, args):
            data[head] = col

        self._data = data
        self.data_keys = {'detector': 'Detector', 'monitor': 'Monitor', 'time': 'Time'}
        self._file_header = file_header

        if build_hkl:
            self.Q_keys = {'h': 'QX', 'k': 'QY', 'l': 'QZ', 'e': 'E', 'temp': 'Temp'}

        if load_instrument:
            instrument = Instrument()
            self.instrument = instrument



# -*- coding: utf-8 -*-
r"""Physical Models

"""
import numpy as np


def simple_harmonic_oscillator(p, t):
    r"""Standard equation for a simple harmonic oscillator

    Parameters
    ----------
    p : list
        Parameters, in the following format:

            +-------+----------------------------+
            | p[0]  | Constant background        |
            +-------+----------------------------+
            | p[1]  | Amplitude                  |
            +-------+----------------------------+
            | p[2]  | Period                     |
            +-------+----------------------------+
            | p[3]  | Phase                      |
            +-------+----------------------------+

    t : ndarray
        One dimensional input array

    Returns
    -------
    func : ndarray
        One dimensional array

    """
    return p[0] + p[1] * np.cos(p[2] * t - p[3])


def damped_harmonic_oscillator(p, t):
    r"""Standard equation for a damped harmonic oscillator

    Parameters
    ----------
    p : list
        Parameters, in the following format:

            +-------+----------------------------+
            | p[0]  | Constant background        |
            +-------+----------------------------+
            | p[1]  | Amplitude                  |
            +-------+----------------------------+
            | p[2]  | Period                     |
            +-------+----------------------------+
            | p[3]  | Phase                      |
            +-------+----------------------------+
            | p[4]  | Damping strength           |
            +-------+----------------------------+

    t : ndarray
        One dimensional input array

    Returns
    -------
    func : ndarray
        One dimensional array

    """
    return p[0] + np.exp(-p[4] * t / 2.) * simple_harmonic_oscillator([0, p[1], p[2], p[3]], t)


def acoustic_phonon_dispersion(p, x):
    r"""Standard equation for the dispersion of an acoustic phonon

    Parameters
    ----------
    p : list
        Parameters, in the following format:

            +-------+----------------------------+
            | p[0]  | Dispersion amplitude       |
            +-------+----------------------------+
            | p[1]  | Dispersion period          |
            +-------+----------------------------+


    x : ndarray
        One dimensional input array

    Returns
    -------
    func : ndarray
        One dimensional array

    """
    return np.sqrt(4 * p[0]) * np.abs(np.sin(p[1] * x))


def optical_phonon_disperion():
    r"""Standard equation for the dispersion of an optical phonon

    """
    pass


def ferromagnetic_disperion(p, x):
    r"""Standard equation for the dispersion of a ferromagnetic spin excitation

    Parameters
    ----------
    p : list
        Parameters, in the following format:

            +-------+----------------------------+
            | p[0]  | Dispersion amplitude       |
            +-------+----------------------------+
            | p[1]  | Dispersion period          |
            +-------+----------------------------+

    x : ndarray
        One dimensional input array

    Returns
    -------
    func : ndarray
        One dimensional array

    """
    return 4 * p[0] * (1 - np.cos(p[1] * x))


def antiferromagnetic_disperion(p, x):
    r"""Standard equation for the dispersion of an antiferromagnetic spin excitation

    Parameters
    ----------
    p : list
        Parameters, in the following format:

            +-------+----------------------------+
            | p[0]  | Dispersion amplitude       |
            +-------+----------------------------+
            | p[1]  | Dispersion period          |
            +-------+----------------------------+

    x : ndarray
        One dimensional input array

    Returns
    -------
    func : ndarray
        One dimensional array

    """
    return p[0] * np.abs(np.sin(p[1] * x))



# -*- coding: utf-8 -*-
r"""Defines Detector class for use in TimeOfFlightInstrument
"""


class Detector(object):
    r"""Class defining a neutron detector for Time of Flight spectrometer
    resolution calculations.

    Parameters
    ----------
    shape : str
        Shape of the detector. 'cylindrical' or 'spherical'

    width : list
        Horizontal coverage of the detector from sample, where 0 is the
        angle of ki where it would hit the detector, in degrees.

    height : list
        Vertical coverage of the detector from the sample, where 0 is the
        position of ki where it would hit the detector, in degrees.

    radius : float
        Radius of curvature of the detector, i.e. the distance
        from the sample to the detector, in cm.

    hpixels : int
        Angular acceptance of a single detector in the horizontal orientation,
        in arc minutes.

    vpixels : int
        Number of detector pixels in the vertical direction

    tau : float, optional
        Binning of the detector in microseconds.

    thickness : float, optional
        Thickness of the detector in cm.

    orientation : string
        Orientation of the cylinder, 'horizontal' or 'vertical', where the
        radius of curvature rotates around the horizontal or vertical axis,
        respectively. Required for shape == 'cylindrical'.

    dead_angles : array-like, optional
        List where dead angles are entered such that [start, stop], in
        degrees. If more than one range of dead angles, pass list of lists.

    Attributes
    ----------
    shape
    width
    height
    radius
    tau
    thickness
    orientation
    dead_angles

    """

    def __init__(self, shape, width, height, radius, hpixels, vpixels, tau=0.1, thickness=1, orientation=None, dead_angles=None):
        self.shape = shape
        self.width = width
        self.height = height
        self.radius = radius
        self.tau = tau
        self.thickness = thickness
        self.hpixels = hpixels
        self.vpixels = vpixels
        if dead_angles:
            self.dead_angles = dead_angles
        if orientation:
            self.orientation = orientation

    def __repr__(self):
        args = ', '.join([str(getattr(self, key)) for key in ['shape', 'width', 'height', 'radius']])
        kwargs = ', '.join(
            ['{0}={1}'.format(key, getattr(self, key, None)) for key in ['resolution', 'orientation', 'dead_angles']])
        return "Detector({0})".format(', '.join([args, kwargs]))


# -*- coding: utf-8 -*-
import numpy as np


def gram_schmidt(in_vecs, row_vecs=True, normalize=True):
    r"""

    """
    if not row_vecs:
        in_vecs = in_vecs.T

    out_vecs = in_vecs[0:1, :]

    for i in range(1, in_vecs.shape[0]):
        proj = np.diag((in_vecs[i, :].dot(out_vecs.T) / np.linalg.norm(out_vecs, axis=1) ** 2).flat).dot(out_vecs)
        out_vecs = np.vstack((out_vecs, in_vecs[i, :] - proj.sum(0)))

    if normalize:
        out_vecs = np.diag(1 / np.linalg.norm(out_vecs, axis=1)).dot(out_vecs)

    if not row_vecs:
        out_vecs = out_vecs.T

    return out_vecs



import numpy as np

from .exceptions import InstrumentError
from .tools import (_CleanArgs, _scalar, calculate_projection_hwhm, ellipse,
                    project_into_plane)


class GeneralInstrument(object):
    r"""Class containing methods general to both Triple Axis and Time of Flight
    instruments.

    Methods
    -------
    calc_projections
    get_resolution_params
    get_resolution

    """

    def calc_projections(self, hkle, npts=36):
        r"""Calculates the resolution ellipses for projections and slices from
        the resolution matrix.

        Parameters
        ----------
        hkle : list
            Positions at which projections should be calculated.

        npts : int, optional
            Number of points in the outputted ellipse curve

        Returns
        -------
        projections : dictionary
            A dictionary containing projections in the planes: QxQy, QxW, and
            QyW, both projections and slices

        """
        R0, NP = self.get_resolution(hkle)

        [H, K, L, W] = _CleanArgs(*hkle)[1:]
        hkle = [H, K, L, W]

        if len(NP.shape) == 2:
            length = 1
        else:
            length = NP.shape[0]

        self.projections = {'QxQy': np.zeros((length, 2, npts)),
                            'QxQySlice': np.zeros((length, 2, npts)),
                            'QxW': np.zeros((length, 2, npts)),
                            'QxWSlice': np.zeros((length, 2, npts)),
                            'QyW': np.zeros((length, 2, npts)),
                            'QyWSlice': np.zeros((length, 2, npts)),
                            'QxQy_fwhm': np.zeros((length, 2)),
                            'QxQySlice_fwhm': np.zeros((length, 2)),
                            'QxW_fwhm': np.zeros((length, 2)),
                            'QxWSlice_fwhm': np.zeros((length, 2)),
                            'QyW_fwhm': np.zeros((length, 2)),
                            'QyWSlice_fwhm': np.zeros((length, 2))}

        A = NP.copy()

        if A.shape == (4, 4):
            A = A.reshape((1, 4, 4))
            R0 = R0[np.newaxis]

        for ind in range(length):
            # Remove the vertical component from the matrix.
            Bmatrix = np.matrix([np.concatenate((A[ind, 0, 0:2], [A[ind, 0, 3]])),
                                 np.concatenate((A[ind, 1, 0:2], [A[ind, 1, 3]])),
                                 np.concatenate((A[ind, 3, 0:2], [A[ind, 3, 3]]))])
            # Projection into Qx, Qy plane
            hwhm_xp, hwhm_yp, theta = calculate_projection_hwhm(project_into_plane(2, R0[ind], Bmatrix)[-1])

            self.projections['QxQy_fwhm'][ind, 0] = 2 * hwhm_xp
            self.projections['QxQy_fwhm'][ind, 1] = 2 * hwhm_yp

            self.projections['QxQy'][ind] = ellipse(hwhm_xp, hwhm_yp, theta,
                                                    [np.dot([self.H[ind], self.K[ind], self.L[ind]],
                                                             self.orient1 / np.linalg.norm(self.orient1) ** 2),
                                                     np.dot([self.H[ind], self.K[ind], self.L[ind]],
                                                             self.orient2 / np.linalg.norm(self.orient2) ** 2)],
                                                    npts=npts)

            # Slice through Qx,Qy plane
            hwhm_xp, hwhm_yp, theta = calculate_projection_hwhm(A[ind, :2, :2])

            self.projections['QxQySlice_fwhm'][ind, 0] = 2 * hwhm_xp
            self.projections['QxQySlice_fwhm'][ind, 1] = 2 * hwhm_yp

            self.projections['QxQySlice'][ind] = ellipse(hwhm_xp, hwhm_yp, theta,
                                                         [np.dot([self.H[ind], self.K[ind], self.L[ind]],
                                                                 self.orient1 /
                                                                 np.linalg.norm(self.orient1) ** 2),
                                                          np.dot([self.H[ind], self.K[ind], self.L[ind]],
                                                                 self.orient2 /
                                                                 np.linalg.norm(self.orient2) ** 2)],
                                                               npts=npts)

            # Projection into Qx, W plane
            hwhm_xp, hwhm_yp, theta = calculate_projection_hwhm(project_into_plane(1, R0, Bmatrix)[-1])

            self.projections['QxW_fwhm'][ind, 0] = 2 * hwhm_xp
            self.projections['QxW_fwhm'][ind, 1] = 2 * hwhm_yp

            self.projections['QxW'][ind] = ellipse(hwhm_xp, hwhm_yp, theta,
                                                   [np.dot([self.H[ind], self.K[ind], self.L[ind]],
                                                           self.orient1 / np.linalg.norm(self.orient1) ** 2),
                                                    self.W[ind]],
                                                   npts=npts)

            # Slice through Qx,W plane
            hwhm_xp, hwhm_yp, theta = calculate_projection_hwhm(
                np.array([[A[ind, 0, 0], A[ind, 0, 3]], [A[ind, 3, 0], A[ind, 3, 3]]]))

            self.projections['QxWSlice_fwhm'][ind, 0] = 2 * hwhm_xp
            self.projections['QxWSlice_fwhm'][ind, 1] = 2 * hwhm_yp

            self.projections['QxWSlice'][ind] = ellipse(hwhm_xp, hwhm_yp, theta,
                                                        [np.dot([self.H[ind], self.K[ind], self.L[ind]],
                                                                self.orient1 / np.linalg.norm(self.orient1) ** 2),
                                                         self.W[ind]],
                                                        npts=npts)

            # Projections into Qy, W plane
            hwhm_xp, hwhm_yp, theta = calculate_projection_hwhm(project_into_plane(0, R0, Bmatrix)[-1])

            self.projections['QyW_fwhm'][ind, 0] = 2 * hwhm_xp
            self.projections['QyW_fwhm'][ind, 1] = 2 * hwhm_yp

            self.projections['QyW'][ind] = ellipse(hwhm_xp, hwhm_yp, theta,
                                                   [np.dot([self.H[ind], self.K[ind], self.L[ind]],
                                                           self.orient2 / np.linalg.norm(self.orient2) ** 2),
                                                    self.W[ind]],
                                                   npts=npts)

            # Slice through Qy,W plane
            hwhm_xp, hwhm_yp, theta = calculate_projection_hwhm(
                np.array([[A[ind, 1, 1], A[ind, 1, 3]], [A[ind, 3, 1], A[ind, 3, 3]]]))

            self.projections['QyWSlice_fwhm'][ind, 0] = 2 * hwhm_xp
            self.projections['QyWSlice_fwhm'][ind, 1] = 2 * hwhm_yp

            self.projections['QyWSlice'][ind] = ellipse(hwhm_xp, hwhm_yp, theta,
                                                        [np.dot([self.H[ind], self.K[ind], self.L[ind]],
                                                                self.orient2 / np.linalg.norm(self.orient2) ** 2),
                                                         self.W[ind]],
                                                        npts=npts)

    def get_resolution_params(self, hkle, plane, mode='project'):
        r"""Returns parameters for the resolution gaussian.

        Parameters
        ----------
        hkle : list of floats
            Position and energy for which parameters should be returned

        plane : 'QxQy' | 'QxQySlice' | 'QxW' | 'QxWSlice' | 'QyW' | 'QyWSlice'
            Two dimensional plane for which parameters should be returned

        mode : 'project' | 'slice'
            Return the projection into or slice through the chosen plane

        Returns
        -------
        tuple : R0, RMxx, RMyy, RMxy
            Parameters for the resolution gaussian

        """

        try:
            A = self.RMS
        except:
            self.calc_resolution(hkle)
            A = self.RMS

        ind = np.where((self.H == hkle[0]) & (self.K == hkle[1]) & (self.L == hkle[2]) & (self.W == hkle[3]))
        print(ind)
        if len(ind[0]) == 0:
            raise InstrumentError('Resolution at provided HKLE has not been calculated.')

        ind = ind[0][0]

        if len(A.shape) != 3:
            A = A.reshape(1, 4, 4)
            selfR0 = self.R0[np.newaxis]

        # Remove the vertical component from the matrix
        Bmatrix = np.vstack((np.hstack((A[ind, 0, :2:1], A[ind, 0, 3])),
                             np.hstack((A[ind, 1, :2:1], A[ind, 1, 3])),
                             np.hstack((A[ind, 3, :2:1], A[ind, 3, 3]))))

        if plane == 'QxQy':
            R0 = np.sqrt(2 * np.pi / Bmatrix[2, 2]) * selfR0[ind]
            if mode == 'project':
                # Projection into Qx, Qy plane
                R0, MP = project_into_plane(2, R0, Bmatrix)
                return R0, MP[0, 0], MP[1, 1], MP[0, 1]
            if mode == 'slice':
                # Slice through Qx,Qy plane
                MP = np.array(A[ind, :2:1, :2:1])
                return R0, MP[0, 0], MP[1, 1], MP[0, 1]

        if plane == 'QxW':
            R0 = np.sqrt(2 * np.pi / Bmatrix[1, 1]) * selfR0[ind]
            if mode == 'project':
                # Projection into Qx, W plane
                R0, MP = project_into_plane(1, R0, Bmatrix)
                return R0, MP[0, 0], MP[1, 1], MP[0, 1]
            if mode == 'slice':
                # Slice through Qx,W plane
                MP = np.array([[A[ind, 0, 0], A[ind, 0, 3]], [A[ind, 3, 0], A[ind, 3, 3]]])
                return R0, MP[0, 0], MP[1, 1], MP[0, 1]

        if plane == 'QyW':
            R0 = np.sqrt(2 * np.pi / Bmatrix[0, 0]) * selfR0[ind]
            if mode == 'project':
                # Projections into Qy, W plane
                R0, MP = project_into_plane(0, R0, Bmatrix)
                return R0, MP[0, 0], MP[1, 1], MP[0, 1]
            if mode == 'slice':
                # Slice through Qy,W plane
                MP = np.array([[A[ind, 1, 1], A[ind, 1, 3]], [A[ind, 3, 1], A[ind, 3, 3]]])
                return R0, MP[0, 0], MP[1, 1], MP[0, 1]

    def get_resolution(self, hkle):
        r"""Returns the resolution matrix and r0 correction at a given Q in rlu

        Parameters
        ----------
        hkle

        Returns
        -------

        """
        [H, K, L, W] = hkle
        try:
            if np.all(H == self.H) and np.all(K == self.K) and np.all(L == self.L) and np.all(W == self.W):
                NP = np.array(self.RMS)
                R0 = self.R0
            else:
                self.calc_resolution(hkle)
                NP = np.array(self.RMS)
                R0 = self.R0
        except AttributeError:
            self.calc_resolution(hkle)
            NP = np.array(self.RMS)
            R0 = self.R0

        return R0, NP



from .atom import Atom, MagneticAtom
from .lattice import Lattice
from .material import Material
from .sample import Sample
from .symmetry import SpaceGroup
from .structure_factors import MagneticFormFactor, MagneticStructureFactor, NuclearStructureFactor
from . import symmetry
from . import tools



r"""NeutronPy: open source python library for neutron scattering data analysis

"""
from __future__ import absolute_import
import sys
import warnings
import pkg_resources

from . import constants
from . import fileio
from . import functions
from . import instrument
from . import models
from . import spurion
from . import scattering
from .crystal import Lattice
from .crystal import Material
from .crystal import Sample
from .crystal import symmetry
from .data import Data
from .energy import Energy
from .instrument import Instrument
from .lsfit import Fitter

try:
    from . import gui
except ImportError:
    warnings.warn('PyQt5 not found, cannot run Resolution GUI')

__version__ = pkg_resources.require("neutronpy")[0].version

if sys.version_info[:2] == (2, 6) or sys.version_info[:2] == (3, 3):
    warnings.warn('Support for Python 2.6 and Python 3.3 is depreciated and will be dropped in neutronpy 1.1.0',
                  DeprecationWarning)



# -*- coding: utf-8 -*-
r""" Class to calculate the energy of a neutron in various common units
"""
import numpy as np
from scipy.constants import h, hbar, k, m_n

from .constants import JOULES_TO_MEV


class Energy(object):
    u"""Class containing the most commonly used properties of a neutron beam
    given some initial input, e.g. energy, wavelength, velocity, wavevector,
    temperature, or frequency. At least one input must be supplied.

    Parameters
    ----------
    energy : float
        Neutron energy in millielectron volts (meV)
    wavelength : float
        Neutron wavelength in angstroms (Å)
    velocity : float
        Neutron velocity in meters per second (m/s)
    wavevector : float
        Neutron wavevector k in inverse angstroms (1/Å)
    temperature : float
        Neutron temperature in kelvin (K)
    frequency : float
        Neutron frequency in terahertz (THz)

    Returns
    -------
    Energy object
        The energy object containing the properties of the neutron beam

    Attributes
    ----------
    energy
    wavelength
    wavevector
    velocity
    temperature
    frequency
    values
    """

    def __init__(self, energy=None, wavelength=None, velocity=None, wavevector=None, temperature=None, frequency=None):
        self._update_values(energy, wavelength, velocity, wavevector, temperature, frequency)

    def __str__(self):
        return self.values

    def __repr__(self):
        return "Energy({0})".format(self.energy)

    def __eq__(self, right):
        return abs(self.energy - right.energy) < 1e-6

    def __ne__(self, right):
        return not self.__eq__(right)

    def _update_values(self, energy=None, wavelength=None, velocity=None, wavevector=None, temperature=None,
                       frequency=None):
        try:
            if energy is None:
                if wavelength is not None:
                    self.en = h ** 2. / (2. * m_n * (wavelength / 1.0e10) ** 2.) * JOULES_TO_MEV
                elif velocity is not None:
                    self.en = 1. / 2. * m_n * velocity ** 2 * JOULES_TO_MEV
                elif wavevector is not None:
                    self.en = (h ** 2 / (2. * m_n * ((2. * np.pi / wavevector) / 1.e10) ** 2) * JOULES_TO_MEV)
                elif temperature is not None:
                    self.en = k * temperature * JOULES_TO_MEV
                elif frequency is not None:
                    self.en = (hbar * frequency * 2. * np.pi * JOULES_TO_MEV * 1.e12)
            else:
                self.en = energy

            if np.any(self.energy == 0.0):
                if isinstance(self.energy, np.ndarray):
                    self.wavelen = np.full(self.energy.shape, np.nan)
                    self.wavevec = np.zeros(self.energy.shape)

                    self.wavelen[self.energy != 0.0] = np.sqrt(h ** 2 / (2. * m_n * self.energy[self.energy !=0] / JOULES_TO_MEV)) * 1.e10
                    self.wavevec[self.energy != 0.0] = 2. * np.pi / self.wavevec[self.energy != 0.0]
                else:
                    self.wavelen = np.nan
                    self.wavevec = 0.0
            else:
                self.wavelen = np.sqrt(h ** 2 / (2. * m_n * self.energy / JOULES_TO_MEV)) * 1.e10
                self.wavevec = 2. * np.pi / self.wavelength

            self.vel = np.sqrt(2. * self.energy / JOULES_TO_MEV / m_n)
            self.temp = self.energy / k / JOULES_TO_MEV
            self.freq = (self.energy / JOULES_TO_MEV / hbar / 2. / np.pi / 1.e12)

        except AttributeError:
            raise AttributeError("""You must define at least one of the \
                                    following: energy, wavelength, velocity, \
                                    wavevector, temperature, frequency""")

    @property
    def energy(self):
        r"""Energy of the neutron in meV"""
        return self.en

    @energy.setter
    def energy(self, value):
        self._update_values(energy=value)

    @property
    def wavelength(self):
        r"""Wavelength of the neutron in Å"""
        return self.wavelen

    @wavelength.setter
    def wavelength(self, value):
        self._update_values(wavelength=value)

    @property
    def wavevector(self):
        u"""Wavevector k of the neutron in 1/Å"""
        return self.wavevec

    @wavevector.setter
    def wavevector(self, value):
        self._update_values(wavevector=value)

    @property
    def temperature(self):
        r"""Temperature of the neutron in Kelvin"""
        return self.temp

    @temperature.setter
    def temperature(self, value):
        self._update_values(temperature=value)

    @property
    def frequency(self):
        r"""Frequency of the neutron in THz"""
        return self.freq

    @frequency.setter
    def frequency(self, value):
        self._update_values(frequency=value)

    @property
    def velocity(self):
        r"""Velocity of the neutron in m/s"""
        return self.vel

    @velocity.setter
    def velocity(self, value):
        self._update_values(velocity=value)

    @property
    def values(self):
        r"""Prints all of the properties of the Neutron beam

        Parameters
        ----------
        None

        Returns
        -------
        values : string
            A string containing all the properties of the neutron including
            respective units
        """
        values = [u'',
                  u'Energy: {0:3.3f} meV'.format(self.energy),
                  u'Wavelength: {0:3.3f} Å'.format(self.wavelength),
                  u'Wavevector: {0:3.3f} 1/Å'.format(self.wavevector),
                  u'Velocity: {0:3.3f} m/s'.format(self.velocity),
                  u'Temperature: {0:3.3f} K'.format(self.temperature),
                  u'Frequency: {0:3.3f} THz'.format(self.frequency),
                  u'']

        return '\n'.join(values)



# -*- coding: utf-8 -*-
r"""GUI for resolution calculations

TESTING ONLY FOR NOW
"""
import os
import sys

import numpy as np
from matplotlib.backends.backend_qt5agg import \
    FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from PyQt5 import uic
from PyQt5.QtWidgets import (QApplication, QMainWindow, QSizePolicy, QTextEdit,
                             QVBoxLayout)

from .energy import Energy
from .instrument import GetTau, Instrument
from .instrument.plot import PlotInstrument


class MyMplCanvas(FigureCanvas):
    def __init__(self, parent=None, width=261, height=201, dpi=100, qslice='QxQy', projections=dict(), u=[1, 0, 0],
                 v=[0, 1, 0]):
        self.fig = Figure(figsize=(width, height), dpi=dpi, facecolor='w', edgecolor='k')
        self.fig.subplots_adjust(bottom=0.25, left=0.25)

        self.axes = self.fig.add_subplot(111)
        self.axes.hold(True)

        self.compute_initial_figure(self.axes, qslice, projections, u, v)

        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)

        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)

    def compute_initial_figure(self, qslice, projections, u, v):
        pass


class MyStaticMplCanvas(MyMplCanvas, PlotInstrument):
    def __init__(self, *args, **kwargs):
        super(MyStaticMplCanvas, self).__init__(*args, **kwargs)

    def compute_initial_figure(self, axis, qslice, projections, u, v):
        self.plot_slice(axis, qslice, projections, u, v)


class MainWindow(QMainWindow):
    r"""Main Window of Resolution Calculator

    """

    def __init__(self, parent=None):
        super(MainWindow, self).__init__(parent)

        uic.loadUi(os.path.join(os.path.dirname(__file__), 'ui/resolution.ui'), self)

        self.qxqyplot = QVBoxLayout(self.qx_qy_plot_widget)
        self.qxwplot = QVBoxLayout(self.qx_w_plot_widget)
        self.qywplot = QVBoxLayout(self.qy_w_plot_widget)

        self.text_output.setFontPointSize(10)
        self.text_output.setLineWrapMode(QTextEdit.NoWrap)

        self.dir_dict = {'Clockwise': 1, 'Counter-Clockwise': -1}
        self.infin_dict = {'ki': 1, 'kf': -1}
        self.edrop_dict = {'energy (meV)': float(self.energy_input.text()),
                           'wavelength (A)': Energy(wavelength=float(self.energy_input.text())).energy,
                           'wave vector (A-1)': Energy(wavevector=float(self.energy_input.text())).energy}
        self.method_dict = {'Cooper-Nathans': 0, 'Popovici': 1}
        self.moncor_dict = {'On': 1, 'Off': 0}

        self.load_instrument()

        self.load_signals()

    def load_instrument(self):
        self.instrument = Instrument()

        self.instrument.sample.a, self.instrument.sample.b, self.instrument.sample.c = [float(i) for i in
                                                                                        self.abc_input.text().split(
                                                                                            ',')]
        self.instrument.sample.alpha, self.instrument.sample.beta, self.instrument.sample.gamma = [float(i) for i in
                                                                                                   self.abg_input.text().split(
                                                                                                       ',')]
        self.instrument.sample.height = float(self.sample_height_input.text())
        self.instrument.sample.width = float(self.sample_width_input.text())
        self.instrument.sample.depth = float(self.sample_depth_input.text())
        self.instrument.sample.u = [float(i) for i in self.sample_u_input.text().split(',')]
        self.instrument.sample.v = [float(i) for i in self.sample_v_input.text().split(',')]
        self.instrument.sample.dir = self.dir_dict[self.sample_dir_select.currentText()]
        self.instrument.sample.mosaic = float(self.sample_mosaic_input.text())
        self.instrument.sample.vmosaic = float(self.sample_vmosaic_input.text())
        self.instrument.sample.shape_type = self.sample_shape_dropdown.currentText().lower()

        if self.mono_select_dropdown.currentText() == 'Custom':
            self.instrument.mono.tau = 2 * np.pi / float(self.mono_select_input)
        else:
            self.instrument.mono.tau = GetTau(self.mono_select_dropdown.currentText())
            self.mono_select_input.setText(
                '{0:.3f}'.format(2. * np.pi / GetTau(self.mono_select_dropdown.currentText())))
        self.instrument.mono.mosaic = float(self.mono_mosaic_input.text())
        self.instrument.mono.vmosaic = float(self.mono_vmosaic_input.text())
        self.instrument.mono.dir = self.dir_dict[self.mono_dir_select.currentText()]
        self.instrument.mono.height = float(self.mono_height_input.text())
        self.instrument.mono.width = float(self.mono_width_input.text())
        self.instrument.mono.depth = float(self.mono_depth_input.text())

        if self.ana_select_dropdown.currentText() == 'Custom':
            self.instrument.ana.tau = 2 * np.pi / float(self.ana_select_input)
        else:
            self.instrument.ana.tau = GetTau(self.ana_select_dropdown.currentText())
            self.ana_select_input.setText('{0:.3f}'.format(2. * np.pi / GetTau(self.ana_select_dropdown.currentText())))
        self.instrument.ana.mosaic = float(self.ana_mosaic_input.text())
        self.instrument.ana.vmosaic = float(self.ana_vmosaic_input.text())
        self.instrument.ana.dir = self.dir_dict[self.ana_dir_select.currentText()]
        self.instrument.ana.height = float(self.ana_height_input.text())
        self.instrument.ana.width = float(self.ana_width_input.text())
        self.instrument.ana.depth = float(self.ana_depth_input.text())

        self.instrument.efixed = self.edrop_dict[self.energy_dropdown.currentText()]
        self.instrument.infin = self.infin_dict[self.fixed_kikf_dropdown.currentText()]
        self.instrument.hcol = [float(i) for i in self.hcols_input.text().split(',')]
        self.instrument.vcol = [float(i) for i in self.vcols_input.text().split(',')]
        self.instrument.arms = [float(i) for i in self.arms_input.text().split(',')]

        self.instrument.guide.height = float(self.guide_height_input.text())
        self.instrument.guide.width = float(self.guide_width_input.text())

        self.instrument.detector.height = float(self.detector_height_input.text())
        self.instrument.detector.width = float(self.detector_width_input.text())

        self.instrument.moncor = self.moncor_dict[self.moncor_dropdown.currentText()]
        self.instrument.method = self.method_dict[self.method_dropdown.currentText()]

        if self.mono_hcurve_input.text() != 'None':
            self.instrument.mono.rh = float(self.mono_hcurve_input.text())
        if self.mono_vcurve_input.text() != 'None':
            self.instrument.mono.rv = float(self.mono_vcurve_input.text())

        if self.ana_hcurve_input.text() != 'None':
            self.instrument.ana.rh = float(self.ana_hcurve_input.text())
        if self.mono_vcurve_input.text() != 'None':
            self.instrument.ana.rh = float(self.ana_vcurve_input.text())

        self.q = [float(i) for i in self.q_input.text().split(',')]

        self.instrument.calc_resolution(self.q)
        self.instrument.calc_projections(self.q)

        # TEST CODE !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        try:
            self.clearLayout(self.qxqyplot)
            self.clearLayout(self.qxwplot)
            self.clearLayout(self.qywplot)
        except:
            pass

        qxqy = MyStaticMplCanvas(self.qx_qy_plot_widget, width=261, height=201, dpi=100, qslice='QxQy',
                                 projections=self.instrument.projections, u=self.instrument.orient1,
                                 v=self.instrument.orient2)
        self.qxqyplot.addWidget(qxqy)

        qxw = MyStaticMplCanvas(self.qx_w_plot_widget, width=261, height=201, dpi=100, qslice='QxW',
                                projections=self.instrument.projections, u=self.instrument.orient1,
                                v=self.instrument.orient2)
        self.qxwplot.addWidget(qxw)

        qyw = MyStaticMplCanvas(self.qy_w_plot_widget, width=261, height=201, dpi=100, qslice='QyW',
                                projections=self.instrument.projections, u=self.instrument.orient1,
                                v=self.instrument.orient2)
        self.qywplot.addWidget(qyw)

        self.text_output.setText(self.instrument.description_string())

    def load_signals(self):
        self.method_dropdown.currentIndexChanged.connect(self.load_instrument)
        self.mono_dir_select.currentIndexChanged.connect(self.load_instrument)
        self.sample_dir_select.currentIndexChanged.connect(self.load_instrument)
        self.ana_dir_select.currentIndexChanged.connect(self.load_instrument)
        self.mono_select_dropdown.currentIndexChanged.connect(self.load_instrument)
        self.ana_select_dropdown.currentIndexChanged.connect(self.load_instrument)
        self.moncor_dropdown.currentIndexChanged.connect(self.load_instrument)
        self.fixed_kikf_dropdown.currentIndexChanged.connect(self.load_instrument)

        self.energy_input.editingFinished.connect(self.load_instrument)
        self.mono_select_input.editingFinished.connect(self.load_instrument)
        self.mono_mosaic_input.editingFinished.connect(self.load_instrument)
        self.mono_vmosaic_input.editingFinished.connect(self.load_instrument)
        self.mono_height_input.editingFinished.connect(self.load_instrument)
        self.mono_width_input.editingFinished.connect(self.load_instrument)
        self.mono_depth_input.editingFinished.connect(self.load_instrument)
        self.mono_hcurve_input.editingFinished.connect(self.load_instrument)
        self.mono_vcurve_input.editingFinished.connect(self.load_instrument)

        self.ana_select_input.editingFinished.connect(self.load_instrument)
        self.ana_mosaic_input.editingFinished.connect(self.load_instrument)
        self.ana_vmosaic_input.editingFinished.connect(self.load_instrument)
        self.ana_height_input.editingFinished.connect(self.load_instrument)
        self.ana_width_input.editingFinished.connect(self.load_instrument)
        self.ana_depth_input.editingFinished.connect(self.load_instrument)
        self.ana_hcurve_input.editingFinished.connect(self.load_instrument)
        self.ana_vcurve_input.editingFinished.connect(self.load_instrument)

        self.abc_input.editingFinished.connect(self.load_instrument)
        self.abg_input.editingFinished.connect(self.load_instrument)
        self.sample_mosaic_input.editingFinished.connect(self.load_instrument)
        self.sample_vmosaic_input.editingFinished.connect(self.load_instrument)
        self.sample_height_input.editingFinished.connect(self.load_instrument)
        self.sample_width_input.editingFinished.connect(self.load_instrument)
        self.sample_depth_input.editingFinished.connect(self.load_instrument)
        self.sample_u_input.editingFinished.connect(self.load_instrument)
        self.sample_v_input.editingFinished.connect(self.load_instrument)
        self.sample_shape_dropdown.currentIndexChanged.connect(self.load_instrument)

        self.hcols_input.editingFinished.connect(self.load_instrument)
        self.vcols_input.editingFinished.connect(self.load_instrument)
        self.arms_input.editingFinished.connect(self.load_instrument)

        self.guide_height_input.editingFinished.connect(self.load_instrument)
        self.guide_width_input.editingFinished.connect(self.load_instrument)
        self.detector_height_input.editingFinished.connect(self.load_instrument)
        self.detector_width_input.editingFinished.connect(self.load_instrument)

        self.energy_input.returnPressed.connect(self.load_instrument)

    def clearLayout(self, layout):
        while layout.count():
            child = layout.takeAt(0)
            child.widget().deleteLater()


def launch():
    app = QApplication(sys.argv)

    w = MainWindow()
    w.show()

    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
import copy
import numbers

import numpy as np

from .instrument import save_instrument
from .loaders import DcsMslice, Grasp, Ice, Icp, Mad, Neutronpy, Spice


def load_data(files, filetype='auto', tols=1e-4, build_hkl=True, load_instrument=False):
    r"""Loads one or more files and creates a :class:`Data` object with the
    loaded data.

    Parameters
    ----------
    files : str or tuple of str
        A file or non-keyworded list of files containing data for input.

    filetype : str, optional
        Default: `'auto'`. Specify file type; Currently supported file types
        are `'SPICE'` (HFIR), `'ICE'` and `'ICP'` (NIST), `'MAD'` (ILL),
        `'dcs_mslice'` DAVE exported ascii formats, GRASP exported ascii and
        HDF5 formats, and neutronpy exported formats. By default the function
        will attempt to determine the filetype automatically.

    tols : float or array_like
        Default: `1e-4`. A float or array of shape `(5,)` giving tolerances
        for combining multiple files. If multiple points are within the given
        tolerances then they will be combined into a single point. If a float
        is given, tolerances will all be the same for all variables in **Q**.
        If an array is given tolerances should be in the format
        `[h, k, l, e, temp]`.

    Returns
    -------
    Data : object
        A :class:`Data` object populated with the data from the input file or
        files.

    """
    load_filetype = {'dcs_mslice': DcsMslice,
                     'grasp': Grasp,
                     'ice': Ice,
                     'icp': Icp,
                     'mad': Mad,
                     'neutronpy': Neutronpy,
                     'spice': Spice}

    if not isinstance(files, (tuple, list)):
        files = (files,)

    for filename in files:
        if filetype == 'auto':
            try:
                filetype = detect_filetype(filename)
            except ValueError:
                raise

        try:
            _data_object_temp = load_filetype[filetype.lower()]()
            _data_object_temp.load(filename, build_hkl=build_hkl, load_instrument=load_instrument)
        except KeyError:
            raise KeyError('Filetype not supported.')

        if isinstance(tols, numbers.Number):
            tols = [tols for i in range(len(_data_object_temp._data) - len(_data_object_temp.data_keys))]

        try:
            _data_object.combine_data(_data_object_temp, tols=tols)
        except (NameError, UnboundLocalError):
            _data_object = copy.deepcopy(_data_object_temp)

    return _data_object


def save_data(obj, filename, filetype='ascii', save_instr=False, overwrite=False, **kwargs):
    """Saves a given object to a file in a specified format.

    Parameters
    ----------
    obj : :class:`Data`
        A :class:`Data` object to be saved to disk

    filename : str
        Path to file where data will be saved

    filetype : str, optional
        Default: `'ascii'`. Data can either be saved in human-readable
        `'ascii'` format, binary `'hdf5'` format, or binary `'pickle'`
        format (not recommended).

    save_instr : bool, optional
        Default: False.

    overwrite : bool, optional
        Default: False.

    """
    if filetype == 'ascii':
        from datetime import datetime

        if overwrite:
            mode = 'w+'
        else:
            mode = 'r+'

        header = '### NeutronPy ::: {0} ###\n\n'.format(datetime.now().isoformat())
        header += 'data_keys = {0}\n'.format(str(obj.data_keys))
        if hasattr(obj, 'Q_keys'):
            header += 'Q_keys = {0}\n'.format(str(obj.Q_keys))

        header += '\n\noriginal_header = \n\t'

        if hasattr(obj, 'file_header'):
            old_header = '\n\t'.join(obj.file_header)
        else:
            old_header = ''

        old_header += '\n\n'

        data_columns = obj.data_columns
        data = obj.data
        if hasattr(obj, '_err'):
            data_columns.append('error')
            data['error'] = obj._err

        col_header = '\nnpy_col_headers =\n' + '\t'.join(data_columns)
        header += old_header + col_header

        output = np.vstack((value for value in data.values())).T

        np.savetxt(filename + '.npy', output, header=header, **kwargs)

        if save_instr:
            save_instrument(obj.instrument, filename, filetype='ascii', overwrite=overwrite)

    elif filetype == 'hdf5':
        import h5py

        if overwrite:
            mode = 'w'
        else:
            mode = 'a'

        with h5py.File(filename + '.h5', mode) as f:
            data = f.create_group('data')

            try:
                data.attrs.create('file_header', obj.file_header.encode('utf8'))
            except AttributeError:
                pass

            data_keys = data.create_group('data_keys')
            for key, value in obj.data_keys.items():
                data_keys.attrs.create(key, value.encode('utf8'))

            if hasattr(obj, 'Q_keys'):
                Q_keys = data.create_group('Q_keys')
                for key, value in obj.Q_keys.items():
                    Q_keys.attrs.create(key, value.encode('utf8'))

            for key, value in obj.data.items():
                data.create_dataset(key, data=value)

            if hasattr(obj, '_err'):
                data.create_dataset('error', data=obj._err)

        if save_instr:
            try:
                save_instrument(obj.instrument, filename, filetype='hdf5', overwrite=False)
            except AttributeError:
                pass

    elif filetype == 'pickle':
        import pickle
        with open(filename, 'wb') as f:
            pickle.dump(obj, f)
    else:
        raise ValueError("""Format not supported. Please use 'ascii', 'hdf5', or 'pickle'""")


def detect_filetype(filename):
    r"""Simple method for quickly determining filetype of a given input file.

    Parameters
    ----------
    filename : str
        File path

    Returns
    -------
    filetype : str
        The filetype of the given input file

    """
    if filename[-3:].lower() == 'nxs':
        return 'grasp'
    elif (filename[-4:].lower() == 'iexy') or (filename[-3:].lower() == 'spe') or (filename[-3:].lower() == 'xye') or (filename[-4:] == 'xyie'):
        return 'dcs_mslice'
    elif filename[-2:].lower() == 'h5' or filename[-3:].lower() == 'npy':
        return 'neutronpy'
    else:
        with open(filename) as f:
            first_line = f.readline()
            second_line = f.readline()
            if '#ICE' in first_line:
                return 'ice'
            elif '# scan' in first_line:
                return 'spice'
            elif 'GRASP' in first_line.upper():
                return 'grasp'
            elif 'Filename' in second_line:
                return 'icp'
            elif 'RRR' in first_line or 'AAA' in first_line or 'VVV' in first_line:
                return 'mad'
            elif 'NeutronPy' in first_line:
                return 'neutronpy'
            else:
                raise ValueError('Unknown filetype.')



import numpy as np

class Guide(object):
    r"""Class for defining a neutron guide object for use with Time of Flight
    spectrometer resolution calculations.

    Attributes
    ----------
    sigma_l
    sigma_theta
    sigma_phi

    """

    def __init__(self, m, length, width, height):
        self.m = m
        self.length = length
        self.width = width
        self.height = height

    @property
    def sigma_l(self):
        return np.sqrt(self.length ** 2 + self.width ** 2 + self.height ** 2) - self.length

    @property
    def sigma_theta(self):
        return np.arcsin(np.sqrt(self.m * np.sin(2.03e-3) ** 2))

    @property
    def sigma_phi(self):
        return self.sigma_theta



from collections import OrderedDict

import numpy as np

from ...data import Data
from ...instrument import Instrument


class Mad(Data):
    r"""Loads MAD (ILL) format ascii data file into a Data object

    """

    def __init__(self):
        super(Mad, self).__init__()

    def load(self, filename, build_hkl=True, load_instrument=False):
        r"""Loads the MAD (ILL) format ascii data file.

        Parameters
        ----------
        filename : str
            Path to file to open

        build_hkl : bool, optional
            Option to build Q = [h, k, l, e, temp]

        load_instrument : bool, optional
            Option to build Instrument from file header

        """
        with open(filename) as f:
            file_header = []
            for i, line in enumerate(f):
                if 'DATA_:' in line:
                    args = next(f).split()
                    col_headers = [head for head in args]
                    skip_lines = i + 2
                    break

        args = np.genfromtxt(filename, unpack=True, dtype=np.float64, skip_header=skip_lines, skip_footer=1)

        data = OrderedDict()
        for head, col in zip(col_headers, args):
            data[head] = col

        self._data = data
        self._file_header = file_header
        self.data_keys = {'detector': 'CNTS', 'time': 'TIME', 'monitor': 'M1'}

        if build_hkl:
            self.Q_keys = {'h': 'QH', 'k': 'QK', 'l': 'QL', 'e': 'EN', 'temp': 'TT'}

        if load_instrument:
            instrument = Instrument()
            self.instrument = instrument



r"""Classes to deal with the Takin resolution calculation engine

"""
import os
import pipes
import sys

from .tas_instrument import TripleAxisInstrument
from .tof_instrument import TimeOfFlightInstrument


class TakinTripleAxis(TripleAxisInstrument):
    r"""Interface between TripleAxisInstrument and Takin engine
    """

    def __init__(self, *args, **kwargs):
        super(TakinTripleAxis, self).__init__(*args, **kwargs)

    def __repr__(self):
        return "Instrument('tas', engine='takin')"

    def calc_resolution(self, hkle):
        pass

    def calc_projections(self, hkle):
        pass

    def resolution_convolution(self):
        pass


class TakinTimeOfFlight(TimeOfFlightInstrument):
    r"""Interface between TimeOfFlightInstrument and Takin engine
    """

    def __init__(self, *args, **kwargs):
        pass

    def __repr__(self):
        return "Instrument('tof', engine='takin')"

    def calc_resolution(self, hkle):
        pass

    def calc_projections(self, hkle):
        pass

    def resolution_convolution(self):
        pass



# -*- coding: utf-8 -*-
r"""Symmetry operations

"""
import numpy as np

from ..constants import symmetry

space_groups = symmetry()['space_groups']


class SpaceGroup(object):
    r"""Class defining a space group of a crystal

    Attributes
    ----------
    full_name
    generators
    group_number
    hm_symbol
    lattice_type
    string_generators
    symbol
    symmetry_operations
    total_operations

    Methods
    -------
    symmetrize_position

    """
    def __init__(self, symbol='P1'):
        if isinstance(symbol, int):
            for key, value in space_groups.items():
                if value['number'] == symbol:
                    self._symbol = key
        elif isinstance(symbol, str):
            if symbol in space_groups:
                self._symbol = symbol
            else:
                for key, value in space_groups.items():
                    if value['hermann-manguin_symbol'] == symbol or value['full_name'] == symbol:
                        self._symbol = key
        else:
            raise KeyError('{0} is not a valid International symbol, Hermann–Mauguin symbol, or space group number'.format(symbol))

        self.point_group = space_groups[self.symbol]['point_group']
        self.full_name = space_groups[self.symbol]['full_name']
        self._generators_str = space_groups[self.symbol]['generators']
        self.lattice_type = space_groups[self.symbol]['type']
        self.group_number = space_groups[self.symbol]['number']
        self.hm_symbol = space_groups[self.symbol]['hermann-manguin_symbol']
        self._generators_mat = get_generator_from_str(self._generators_str)
        self.total_operations = space_groups[self.symbol]['total_operations']
        self.symmetry_operations = self._symmetry_operations_from_generators()

    def __repr__(self):
        return "SpaceGroup({0})".format(self.group_number)

    @property
    def symbol(self):
        r"""Space group symbol
        """
        return self._symbol

    @symbol.setter
    def symbol(self, symbol):
        self.__init__(symbol)

    @property
    def string_generators(self):
        r"""Space group generators
        """
        return self._generators_str

    @property
    def generators(self):
        r"""Space group generators in matrix format
        """
        return self._generators_mat

    def _symmetry_operations_from_generators(self):
        symm_ops, new_ops = [np.copy(self.generators)] * 2
        while len(new_ops) > 0 and len(symm_ops) < self.total_operations:
            gen_ops = []
            for g in new_ops:
                test_ops = np.einsum('ijk,kl', symm_ops, g)
                for op in test_ops:
                    op[:3, 3] = np.mod(get_translation(op), 1)
                    op[np.where(np.abs(1 - get_translation(op)) < 1e-15), 3] = 0
                    if not (np.abs(symm_ops - op) < 1e-15).all(axis=(1, 2)).any():
                        gen_ops.append(op)
                        symm_ops = np.append(symm_ops, [op], axis=0)
            new_ops = gen_ops
        assert len(symm_ops) == self.total_operations
        return symm_ops

    def symmetrize_position(self, vector):
        r"""Applies symmetry operations to a vector

        """
        positions = []
        for op in self.symmetry_operations:
            positions.append(np.dot(get_rotation(op), np.array(vector)) + get_translation(op))

        return positions


def get_formatted_operations(operations):
    r"""Returns operations formatted in a list for easy parsing

    Parameters
    ----------
    operations

    Returns
    -------

    """
    if (isinstance(operations, list) and isinstance(operations[0], str)) or isinstance(operations, str):
        operations = get_generator_from_str(operations)

    if isinstance(operations, np.ndarray):
        operations = [operations]

    return operations


def get_rotation(operations):
    r"""Returns rotational part of operator

    """
    rotations = []
    for operation in get_formatted_operations(operations):
        rotations.append(operation[:3, :3])

    if len(rotations) == 1:
        rotations = rotations[0]

    return rotations


def get_translation(operations):
    r"""Returns rotational part of operator

    """
    translations = []
    for operation in get_formatted_operations(operations):
        translations.append(operation[:3, 3])

    if len(translations) == 1:
        translations = translations[0]

    return translations


def get_generator_from_str(operations):
    r"""Returns generator arrays

    Returns
    -------
    operators : list of ndarrays
        List of operation arrays with shape (3,4)

    """
    if isinstance(operations, str):
        operations = [operations]

    operators = []
    for gen in operations:
        components = gen.split(',')
        if len(components) > 3:
            raise ValueError('Generator string {0} is in wrong format'.format(gen))

        rotation = np.zeros((3, 3))
        translation = np.zeros(3)
        for i, comp in enumerate(components):
            elements = comp.split('+')
            if len(elements) > 1:
                translation[i] = eval(elements[-1].replace('/', './'))

            if '-x' in elements[0]:
                rotation[i, 0] = -1
            elif 'x' in elements[0]:
                rotation[i, 0] = 1

            if '-y' in elements[0]:
                rotation[i, 1] = -1
            elif 'y' in elements[0]:
                rotation[i, 1] = 1

            if '-z' in elements[0]:
                rotation[i, 2] = -1
            elif 'z' in elements[0]:
                rotation[i, 2] = 1

        out = np.zeros((4, 4), dtype=float)
        out[0:3, 0:3] = rotation
        out[0:3, 3] = translation
        out[3, 3] = 1.

        operators.append(out)

    if len(operators) == 1:
        operators = operators

    return operators


def get_str_from_generator(operations):
    r"""Returns strings of generators from arrays

    Parameters
    ----------
    operations : str, array, list

    Returns
    -------
    generators : list of str
        List of generator strings

    """
    if isinstance(operations, np.ndarray) and len(operations.shape) < 3:
        operations = [operations]

    syms = ['x', 'y', 'z']
    signs = {-1: '-', 1: '+', 0: ''}
    generators = []
    for operation in operations:
        line = []
        for row in operation[:3, :]:
            element = ''
            for col, sym in zip(row[:3], syms):
                element += signs[int(col)] + np.abs(int(col)) * sym

            if row[3] == 0:
                translate = ''
            elif np.round(1. / row[3], 1) == 1.5:
                translate = '2/3'
            elif row[3] == 0.75:
                translate = '3/4'
            elif np.round(row[3], 3) == 0.833:
                translate = '5/6'
            else:
                denominator = int(np.round(1. / row[3]))
                translate = '1/{0}'.format(denominator)

            if translate != '':
                element += '+{0}'.format(translate)

            if len(element) >= 1 and element[0] == '+':
                element = element[1:]

            line.append(element)

        generators.append(','.join(line))

    return generators



# -*- coding: utf-8 -*-
from .monochromator import Monochromator


class Analyzer(Monochromator):
    u"""Class containing analyzer information.

    Parameters
    ----------
    tau : float or str
        Tau value for the analyzer

    mosaic : int
        Mosaic of the analyzer in arc minutes

    direct : ±1, optional
        Direction of the analyzer (left or right, -1 or +1, respectively).
        Default: -1 (left-handed coordinate frame).

    vmosaic : int, optional
        Vertical mosaic of the analyzer in arc minutes. Default: None

    height : float, optional
        Height of the analyzer in cm. Default: None

    width : float, optional
        Width of the analyzer in cm. Default: None

    depth : float, optional
        Depth of the analyzer in cm. Default: None

    horifoc : int, optional
        Set to 1 if horizontally focusing analyzer is used. Default: -1

    thickness : float, optional
        Thickness of Analyzer crystal in cm. Required for analyzer
        reflectivity calculation. Default: None

    Q : float, optional
        Kinematic reflectivity coefficient. Required for analyzer
        reflectivity calculation. Default: None

    Returns
    -------
    Analyzer : object

    Attributes
    ----------
    tau
    mosaic
    vmosaic
    direct
    height
    width
    depth
    rh
    rv
    thickness
    horifoc
    Q
    d

    """

    def __init__(self, tau, mosaic, direct=-1, vmosaic=None, height=None, width=None, depth=None, rh=None, rv=None,
                 horifoc=-1, thickness=None, Q=None):
        super(Analyzer, self).__init__(tau, mosaic, direct=-1, vmosaic=None, height=None, width=None, depth=None,
                                       rh=None, rv=None)
        if thickness is not None:
            self.thickness = thickness
        if Q is not None:
            self.Q = Q
        self.horifoc = horifoc

    def __repr__(self):
        args = ', '.join([str(getattr(self, key)) for key in ['tau', 'mosaic']])
        kwargs = ', '.join(['{0}={1}'.format(key, getattr(self, key)) for key in
                            ['direct', 'vmosaic', 'height', 'width', 'depth', 'rh', 'rv', 'horifoc', 'thickness', 'Q']
                            if getattr(self, key, None) is not None])
        return "Analyzer({0})".format(', '.join([args, kwargs]))



# -*- coding: utf-8 -*-
r"""Define Triple Axis goniometer

"""
import numpy as np


class Goniometer(object):
    r"""Class for defining a goniometer.

    Attributes
    ----------
    omega_rad
    sgu_rad
    sgl_rad
    theta_rad
    theta
    N
    M
    Omega
    Theta
    T_c
    T_Phi
    R
    U

    Methods
    -------
    u_phi

    """

    def __init__(self, u, theta_u, v, theta_v, sgu, sgl, omega=0):
        self.u = u
        self.theta_u = theta_u

        self.v = v
        self.theta_v = theta_v

        self.sgu = sgu
        self.sgl = sgl

        self.omega = omega

    def __repr__(self):
        return "Goniometer({0})".format(
            ', '.join([str(getattr(self, key)) for key in ['u', 'theta_u', 'v', 'theta_v', 'sgu', 'sgl', 'omega']]))

    @property
    def omega_rad(self):
        return self.omega_rad

    @property
    def sgu_rad(self):
        return np.deg2rad(self.sgu)

    @property
    def sgl_rad(self):
        return np.deg2rad(self.sgl)

    @property
    def theta_rad(self):
        return np.arctan((self.ki - self.kf * np.cos(self.phi)) / (self.kf * np.sin(self.phi)))

    @property
    def theta(self):
        pass

    @property
    def N(self):
        return np.matrix([[1, 0, 0],
                          [0, np.cos(self.sgu_rad), -np.sin(self.sgu_rad)],
                          [0, np.sin(self.sgu_rad), np.cos(self.sgu_rad)]])

    @property
    def M(self):
        return np.matrix([[np.cos(self.sgl_rad), 0, np.sin(self.sgl_rad)],
                          [0, 1, 0],
                          [-np.sin(self.sgl_rad), 0, np.cos(self.sgl_rad)]])

    @property
    def Omega(self):
        return np.matrix([[np.cos(self.omega_rad), -np.sin(self.omega_rad), 0],
                          [np.sin(self.omega_rad), np.cos(self.omega_rad), 0],
                          [0, 0, 1]])

    @property
    def Theta(self):
        return np.matrix([[np.cos(self.theta_rad), -np.sin(self.theta_rad), 0],
                          [np.sin(self.theta_rad), np.cos(self.theta_rad), 0],
                          [0, 0, 1]])

    @property
    def T_c(self):
        return np.matrix([self.u, self.v, np.cross(self.u, self.v)]).T

    @property
    def T_phi(self):
        return np.matrix([self.u_phi(np.deg2rad(self.theta_u), self.sgl_rad, self.sgu_rad),
                          self.u_phi(np.deg2rad(self.theta_v), self.sgl_rad, self.sgu_rad),
                          self.u_phi(np.deg2rad(0), np.deg2rad(90), np.deg2rad(0))]).T

    @property
    def R(self):
        return self.Omega * self.M * self.N

    @property
    def U(self):
        r"""Defines an orientation matrix based on supplied goniometer angles

        """
        return self.T_phi * np.linalg.inv(self.T_c)

    def u_phi(self, omega, chi, phi):
        r"""Vector to help convert goniometer angles to lab reference frame orientation

        Parameters
        ----------
        omega : float
            Goniometer rotation axis in rad

        chi : float
            Goniometer upper tilt in rad

        phi : float
            Goniometer lower tilt angle in rad

        Returns
        -------
        rot : array
            Array to help convert between lab reference goniometer and sample reference
        """
        return [np.cos(omega) * np.cos(chi) * np.cos(phi) - np.sin(omega) * np.sin(phi),
                np.cos(omega) * np.cos(chi) * np.sin(phi) + np.sin(omega) * np.cos(phi),
                np.cos(omega) * np.sin(chi)]



# -*- coding: utf-8 -*-
r"""Material constructor

"""
import neutronpy.constants as const
import numpy as np

from .atom import Atom, MagneticAtom
from .plot import PlotMaterial
from .sample import Sample
from .structure_factors import MagneticStructureFactor, NuclearStructureFactor
from .symmetry import SpaceGroup

# from ..scattering.pattern import HKLGenerator


class MagneticUnitCell(Sample):
    r"""Class defining a magnetic unit cell

    """

    def __init__(self, unit_cell):
        if 'chirality' not in unit_cell:
            self.chirality = 0
        if 'ph' not in unit_cell:
            self.phase = 0
        self.atoms = []
        if 'space_group' in unit_cell:
            self.space_group = SpaceGroup(unit_cell['space_group'])
            for atom in unit_cell['atoms']:
                symmetrized_positions = self.space_group.symmetrize_position(atom['pos'])
                for pos in symmetrized_positions:
                    self.atoms.append(MagneticAtom(atom['ion'],
                                                   pos,
                                                   atom['moment'],
                                                   atom['occupancy']))
        else:
            for atom in unit_cell['atoms']:
                self.atoms.append(MagneticAtom(atom['ion'],
                                               atom['pos'],
                                               atom['moment'],
                                               atom['occupancy']))

        self.propagation_vector = unit_cell['propagation_vector']

        a, b, c = unit_cell['lattice']['abc']
        alpha, beta, gamma = unit_cell['lattice']['abg']
        super(MagneticUnitCell, self).__init__(a, b, c, alpha, beta, gamma)

    def __repr__(self):
        return "MagneticUnitCell('{0}')".format(self.propagation_vector)


class Material(Sample, NuclearStructureFactor, MagneticStructureFactor, PlotMaterial):
    r"""Class for the Material being supplied for the structure factor calculation

    Parameters
    ----------
    data : dictionary
        data is a dictionary containing all of the atoms
        and their positions, with optional occupancy
        and variances in positions (dpos), which may be
        used for Debye-Waller factor. This dictionary has
        the format:

        .. code-block:: python

            {'name': str,
             'composition': [{'ion': str,
                              'pos': [float, float, float],
                              'Uiso': float,
                              'Uaniso': matrix(3,3)
                              'occupancy': float}],
             'massNorm': bool,
             'formulaUnits': float,
             'lattice': {'abc': [float, float, float],
                         'abg': [float, float, float]}

    The following are valid options for the data dictionary:

    'name' : str
        Name of the structure

    'composition' : list of dicts
        For each atom in the unit cell, you must provide:

            'ion' : str
                Name of the atom. Needed for mass and scattering length

            'pos' : list of 3 floats
                x, y, z position of the atom within the unit cell in normalized
                units

            'dpos' : list of 3 floats
                x, y, z displacements of the atom, for Debye-Waller factor, in
                normalized units

            'occupancy' : float
                Occupancy of the site, e.g. if atoms only partially occupy this
                site

    'Uiso' : bool
        Include Debye-Waller in calculation with isotropic U

    'Uaniso' : bool
        Include Debye-Waller in calculation with anisotropic U

    'massNorm' : bool
        Normalize calculations to the square-root of the mass of atoms. This
        is useful for the calculation of the coherent one-phonon inelastic
        cross-section, which is dependent on a nuclear structure factor in
        which the nuclear scattering length is normalized by the square-root
        of the mass, *i.e.* :math:`\bar{b}_d/\sqrt{M_d}`, see Eq. 4.88 in
        "Theory of neutron scattering from condensed matter, Volume 1" by
        Stephen W. Lovesey.

    'formulaUnits' : float
        Number of formula units to use in the calculation

    'lattice' : dict

        'abc' : lattice constants of unit cell

        'abg' : lattice angles of unit cell

    'space_group' : str or int
        Hermann–Mauguin symbol or international space group number

    Returns
    -------
    output : object
        Material Object defining a single crystal.

    Attributes
    ----------
    volume
    total_scattering_cross_section
    a
    b
    c
    astar
    bstar
    cstar
    alpha
    beta
    gamma
    alpha_rad
    beta_rad
    gamma_rad
    alphastar
    betastar
    gammastar
    alphastar_rad
    betastar_rad
    gammastar_rad
    abg_rad
    reciprocal_abc
    reciprocal_abg
    reciprocal_abg_rad
    lattice_type
    volume
    reciprocal_volume
    G
    Gstar
    Bmatrix
    Umatrix
    UBmatrix
    u
    v

    Methods
    -------
    calc_nuc_str_fac
    calc_mag_str_fac
    calc_mag_int_vec
    calc_incoh_elas_xs
    calc_optimal_thickness
    plot_unit_cell
    get_angle_between_planes
    get_d_spacing
    get_q
    get_two_theta
    get_phi
    N_atoms
    apply_scattering_rules
    find_equivalent_positions
    find_site_multiplicity
    generate_hkl_positions

    """

    def __init__(self, crystal):
        self.name = crystal['name']

        if 'formulaUnits' not in crystal:
            crystal['formulaUnits'] = 1.

        self.muCell = 0.
        for item in crystal['composition']:
            if 'occupancy' not in item:
                item['occupancy'] = 1.
            self.muCell += const.periodic_table()[item['ion']]['mass'] * item['occupancy']

        self.Mcell = self.muCell * crystal['formulaUnits']

        if 'lattice' in crystal:
            a, b, c = crystal['lattice']['abc']
            alpha, beta, gamma = crystal['lattice']['abg']

        if 'wavelength' in crystal:
            self.wavelength = crystal['wavelength']
        else:
            self.wavelength = 2.359

        if 'space_group' in crystal:
            self.space_group = SpaceGroup(crystal['space_group'])
            self.atoms = []
            for atom in crystal['composition']:
                if 'Uiso' not in atom:
                    atom['Uiso'] = 0
                if 'Uaniso' not in item:
                    atom['Uaniso'] = np.matrix(np.zeros((3, 3)))
                if 'occupancy' not in item:
                    atom['occupancy'] = 1.
                symmetrized_positions = self.space_group.symmetrize_position(atom['pos'])
                for pos in symmetrized_positions:
                    self.atoms.append(Atom(atom['ion'],
                                           pos,
                                           atom['occupancy'],
                                           self.Mcell,
                                           crystal['massNorm'],
                                           atom['Uiso'],
                                           atom['Uaniso']))
        else:
            self.atoms = []
            for item in crystal['composition']:
                if 'Uiso' not in item:
                    item['Uiso'] = 0
                if 'Uaniso' not in item:
                    item['Uaniso'] = np.matrix(np.zeros((3, 3)))
                if 'occupancy' not in item:
                    item['occupancy'] = 1.
                self.atoms.append(Atom(item['ion'],
                                       item['pos'],
                                       item['occupancy'],
                                       self.Mcell,
                                       crystal['massNorm'],
                                       item['Uiso'],
                                       item['Uaniso']))

        if 'magnetic_unit_cell' in crystal:
            self.magnetic_unit_cell = MagneticUnitCell(crystal['magnetic_cell'])

        if 'mosaic' not in crystal:
            crystal['mosaic'] = None
        if 'vmosaic' not in crystal:
            crystal['vmosaic'] = None
        if 'u' not in crystal:
            crystal['u'] = None
        if 'v' not in crystal:
            crystal['v'] = None
        if 'dir' not in crystal:
            crystal['dir'] = 1

        super(Material, self).__init__(a, b, c, alpha, beta, gamma, crystal['mosaic'], crystal['vmosaic'],
                                       crystal['dir'], crystal['u'], crystal['v'])

    def __repr__(self):
        return "Material('{0}')".format(self.name)

    @property
    def total_scattering_cross_section(self):
        r"""Returns total scattering cross-section of unit cell
        """
        total = 0
        for atom in self.atoms:
            total += (atom.coh_xs + atom.inc_xs)
        return total

    def N_atoms(self, mass):
        r"""Number of atoms in the defined Material, given the mass of the sample.

        Parameters
        ----------
        mass : float
            The mass of the sample in grams.

        Returns
        -------
        Natoms : int
            The number of atoms of the material based on the mass of the sample.

        """
        return const.N_A * mass / self.muCell

    def calc_optimal_thickness(self, energy=25.3, transmission=1 / np.exp(1)):
        r"""Calculates the optimal sample thickess to avoid problems with
        extinction, multiple coherent scattering and absorption.

        Parameters
        ----------
        energy : float, optional
            The energy of the incident beam in meV. Default: 25.3 meV.

        transmission: float, optional
            The transmission through the material in decimal percentage,
            :math:`0 < T < 1.0`. Default: :math:`1/e`.

        Returns
        -------
        thickness : float
            Returns the optimal thickness of the sample in cm

        Notes
        -----
        The total transmission of neutrons through a material is defined as

        .. math:: T = \frac{I}{I_0} = e^{-\Sigma_T d},

        where :math:`\Sigma_T` is the total scattering cross-section, *i.e.*

        .. math:: \Sigma_T = \Sigma_{coh} + \Sigma_{inc} + \Sigma_{abs},

        and :math:`d` is the thickness in cm.

        Scattered intensity is thus defined to be

        .. math:: I_s \propto dT\left(\frac{d\Sigma_{coh}}{d\Omega}\right) \propto d e^{-\Sigma_T d}.

        :math:`I_s` is therefore a maximum when :math:`d=1/\Sigma_T`, resulting
        a transmission of approximately :math:`T=37\%`. This is valid when the
        coherent cross-section is much less than the total cross-section,
        *i.e.*

        .. math:: \Sigma_{coh} \ll \Sigma_T \approx \Sigma_{inc} + \Sigma{abs}.

        However, if the coherent cross-section of the material is the dominant
        part of the total scattering cross-section, *i.e.*
        :math:`\Sigma_T \approx \Sigma_{coh}`, then :math:`d = 1/\Sigma_T` is
        too large and there will be a problem with multiple scattering.
        Therefore a transmission of :math:`T\geq90\%` is desirable.

        """

        sigma_coh = np.sum([atom.occupancy * atom.coh_xs for atom in self.atoms])
        sigma_inc = np.sum([atom.occupancy * atom.inc_xs for atom in self.atoms])
        sigma_abs = np.sum([atom.occupancy * atom.abs_xs for atom in self.atoms])

        sigma_T = (sigma_coh + sigma_inc + sigma_abs * np.sqrt(25.3 / energy)) / self.volume

        return -np.log(transmission) / sigma_T

    def calc_incoh_elas_xs(self, mass=None):
        r"""Calculates the incoherent elastic cross section.

        Parameters
        ----------
        mass : float (optional)
            The mass of the sample in grams (Default: None)

        Returns
        -------
        INC_XS : float
            Incoherent Elastic Cross Section

        """

        INC_XS = 0
        for atom in self.atoms:
            INC_XS += atom.inc_xs * np.exp(-8 * np.pi ** 2 * atom.Uiso * np.sin(
                np.deg2rad(self.get_two_theta(atom.pos, self.wavelength) / 2.)) ** 2 / self.wavelength ** 2)

        if mass is not None:
            return self.N_atoms(mass) / (4 * np.pi) * INC_XS
        else:
            return INC_XS / (4 * np.pi)



import numpy as np


class PlotMaterial(object):
    r'''Class containing plotting methods for Material object

    Methods
    -------
    plot_unit_cell

    '''
    def plot_unit_cell(self):
        r'''Plots the unit cell and atoms of the material.

        '''

        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D  # @UnresolvedImport
        from itertools import product, combinations

        fig = plt.figure()
        ax = fig.gca(projection='3d')

        # draw unit cell
        for s, e in combinations(np.array(list(product([0, self.abc[0]], [0, self.abc[1]], [0, self.abc[2]]))), 2):
            if np.sum(np.abs(s - e)) in self.abc:
                ax.plot3D(*zip(s, e), color="b")

        # plot atoms
        x, y, z, m = [], [], [], []
        for item in self.atoms:
            x.append(item.pos[0] * self.abc[0])
            y.append(item.pos[1] * self.abc[1])
            z.append(item.pos[2] * self.abc[2])
            m.append(item.mass)

        ax.scatter(x, y, z, s=m)

        plt.axis('scaled')
        plt.axis('off')

        plt.show()



# -*- coding: utf-8 -*-
from warnings import warn
from .takin import TakinTimeOfFlight, TakinTripleAxis
from .tas_instrument import TripleAxisInstrument
from .tof_instrument import TimeOfFlightInstrument


class Instrument(object):
    r"""An object that represents either a Triple Axis Spectrometer instrument
    or a Time of Flight Instrument configuration, including a sample.

    Parameters
    ----------
    args : arg, optional
        Any valid **positional** arguments for the desired instrument class

    instrument_type : str, optional
        Used to select Triple Axis instrument 'tas' or time of flight
        instrument 'tof'. Default: 'tas'

    engine : str, optional
        Used to select the engine for resolution calculations.
        Default: 'neutronpy' for 'tas' instruments, and 'takin' for 'tof'
        instruments.

    kwargs : kwarg, optional
        Any valid keyword arguments for the desired instrument class

    """

    def __init__(self, *args, **kwargs):
        if 'instrument_type' not in kwargs:
            kwargs['instrument_type'] = 'tas'

        if 'engine' not in kwargs:
            kwargs['engine'] = 'neutronpy'

        self.instrument_type = kwargs['instrument_type']
        self.engine = kwargs['engine']

        if kwargs['instrument_type'] == 'tas':
            if kwargs['engine'] == 'neutronpy':
                self.__class__ = TripleAxisInstrument
                self.__init__(*args, **kwargs)
            elif kwargs['engine'] == 'takin':
                warn("Takin engine is not yet supported")
                self.__class__ = TakinTripleAxis
                self.__init__(*args, **kwargs)

        elif kwargs['instrument_type'] == 'tof':
            if kwargs['engine'] == 'neutronpy':
                self.__class__ = TimeOfFlightInstrument
                self.__init__(*args, **kwargs)

            elif kwargs['engine'] == 'takin':
                warn("Takin engine is not yet supported")
                self.__class__ = TakinTimeOfFlight
                self.__init__(*args, **kwargs)

    def __repr__(self):
        return "Instrument('{0}', engine='{1}')".format(self.instrument_type, self.engine)



# -*- coding: utf-8 -*-
r"""Structure Factors

NuclearStructureFactor
MagneticStructureFactor
MagneticFormFactor

"""
import numpy as np

from ..constants import magnetic_ion_j


class NuclearStructureFactor(object):
    r"""Class containing nuclear structure factor calculator

    Methods
    -------
    calc_nuc_str_fac

    """
    def calc_nuc_str_fac(self, hkl):
        r"""Calculates the structural form factor of the material.

        Parameters
        ----------
        hkl : tuple of floats, or tuple of array-like
            Reciprocal lattice positions at which the structure
            factor should be calculated

        Returns
        -------
         NSF : float or ndarray
             Nuclear structure factor at the position or positions specified

        Notes
        -----

        """

        h, k, l = hkl

        # Ensures input arrays are complex ndarrays
        if isinstance(h, (np.ndarray, list, tuple)):
            h = np.array(h).astype(complex)
        if isinstance(k, (np.ndarray, list, tuple)):
            k = np.array(k).astype(complex)
        if isinstance(l, (np.ndarray, list, tuple)):
            l = np.array(l).astype(complex)

        # construct structure factor
        NSF = 0 * 1j
        for atom in self.atoms:
            NSF += atom.occupancy * atom.b * np.exp(1j * 2. * np.pi * (h * atom.pos[0] + k * atom.pos[1] + l * atom.pos[2])) * \
                np.exp(-8 * np.pi ** 2 * atom.Uiso * np.sin(np.deg2rad(self.get_two_theta(atom.pos, self.wavelength) / 2.)) ** 2 / self.wavelength ** 2) * \
                np.exp(-np.float(np.dot(np.dot(atom.pos, atom.Uaniso), atom.pos)))

        return NSF


class MagneticFormFactor(object):
    r"""Class defining a magnetic ion.

    Parameters
    ----------
    ion : str
        Name of the atom, ion or anion, *e.g.* 'Fe2+'.

    Returns
    -------
    output : Object
        Ion object defining a single magnetic ion.

    Methods
    -------
    calc_mag_form_fac
    """

    def __init__(self, ion):
        self.ion = ion
        try:
            self.j0 = magnetic_ion_j()[self.ion]['j0']
            self.j2 = magnetic_ion_j()[self.ion]['j2']
            self.j4 = magnetic_ion_j()[self.ion]['j4']
        except ValueError:
            raise ValueError('No such ion was found in database.')

    def __repr__(self):
        return "MagneticFormFactor('{0}')".format(self.ion)

    def calc_mag_form_fac(self, q=None, g=None, qrange=None):
        r"""Calculate the magnetic form factor of an ion.

        Parameters
        ----------
        q : float or list, optional
            An array of values or position at which the form
            factor should be calcuated.

        g : float, optional
            The g-factor, which is 2 is left undefined.

        qrange : float, optional
            The range of q over which the form factor should be
            calculated, if no input array q is provided.

        Returns
        -------
        output : tuple
            (form factor, q, j\ :sub:`0`, j\ :sub:`2`, j\ :sub:`4`)

        Notes
        -----
        The magnetic form factor of an ion is given by:

        .. math:: f(q) = <j_0(q)> + (\frac{2}{g}-1)<j_2(q)> \mathrm{(Jensen and Mackintosh,1991)}

        using the 3-gaussian approximation to :math:`f(q)` from the
        International Tables of Crystallography (by J. Brown)


        """

        if q is None:
            if qrange is None:
                q = np.linspace(0., 2., 2. / 0.025 + 1)
            else:
                q = np.linspace(qrange[0], qrange[1], (qrange[1] - qrange[0]) / 0.025 + 1)
        if g is None:
            g = 2.

        x = q / 4. / np.pi

        j0 = (self.j0[0] * np.exp(-self.j0[1] * x ** 2) + self.j0[2] *
              np.exp(-self.j0[3] * x ** 2) + self.j0[4] *
              np.exp(-self.j0[5] * x ** 2) + self.j0[6])

        j2 = x ** 2 * (self.j2[0] * np.exp(-self.j2[1] * x ** 2) +
                       self.j2[2] * np.exp(-self.j2[3] * x ** 2) +
                       self.j2[4] * np.exp(-self.j2[5] * x ** 2) +
                       self.j2[6])

        j4 = x ** 2 * (self.j4[0] * np.exp(-self.j4[1] * x ** 2) +
                       self.j4[2] * np.exp(-self.j4[3] * x ** 2) +
                       self.j4[4] * np.exp(-self.j4[5] * x ** 2) +
                       self.j4[6])

        ff = j0 + (2. / g - 1.) * j2

        return ff, q, j0, j2, j4


class MagneticStructureFactor(object):
    r"""Class containing magnetic structure factor calculator

    Methods
    -------
    calc_mag_int_vec
    calc_mag_str_fac

    """
    def calc_mag_int_vec(self):
        r"""Calculates magnetic interaction vector
        """
        pass

    def calc_mag_str_fac(self):
        r"""Calculates magnetic structure factor
        """
        pass



# -*- coding: utf-8 -*-
r"""A class for handeling a collection of scans

"""
import collections

import matplotlib.colors as mpl_c
import matplotlib.pyplot as plt
import numpy as np


class Scans(object):
    r"""A class for a collection of scans

    Attributes
    ----------
    scans : Ordered dict
        A dictionary of scan objects
    num_scans : int
        The number of scans

    Methods
    -------
    waterfall
    pcolor
    func_col
    min_col
    max_col
    mean_col
    update
    scans_check

    """

    def __init__(self, scans_dict=None):
        if scans_dict is not None:
            if not isinstance(scans_dict, collections.OrderedDict):
                raise RuntimeError(
                    "the input dictionary must be of type OrderedDict")
            self.num_scans = len(scans_dict)
        self.scans = scans_dict

    def update(self, scans_dict):
        r"""Update the scans_dict to include the dictionary scans_dict
        This will update any scans that are already in the class and will
        append those that are not.

        Parameters
        ----------
        scans_dict : dict
              A dictionary of multiple scans to add to the collection.

        """
        if not isinstance(scans_dict, collections.OrderedDict):
            raise RuntimeError(
                "the input dictionary must be of type OrderedDict")
        self.scans.update(scans_dict)
        self.num_scans = len(scans_dict)

    def scans_check(self):
        r"""Check to see if their are scans in the object.

        This should be used for other methods that work on individual scans in
        the collection.

        Raises
        ------
        RuntimeError
             If there are no scans

        """
        if self.scans is None:
            raise RuntimeError('There must be at lest one scan')

    def waterfall(self, x='e', y='detector', label_column='h', offset=5, fmt='b-', legend=False, show_plot=False):
        r"""Create a waterfall plot of all the scans in the collection

        Parameters
        ----------
        x : str, optional
            Name of one of the columns that is in each scan to plot along the
            x-axis. Default: `'e'`.

        y : str, optional
            Name of one of the columns that is in each scan to plot along the
            y-axis. Default: `'detector'`.

        label_column : str, optional
            Name of one of the columns that is in each scan to label each
            separate scan in the plot. Default: `'h'`.

        offset : float, optional
            Offset in y, of each successive curve from the previous one.
            Default: 5.

        fmt : str, optional
            Matplotlib plot format string.  Default: `'b-'`.

        legend : bool, optional
            Flag to plot a legend, where True plots a legend. Default: False.

        """
        self.scans_check()
        fh = plt.figure()
        plt.hold(True)

        for idx, scan_num in enumerate(self.scans.keys()):
            xin = self.scans[scan_num].data[x]
            yin = self.scans[scan_num].data[y]
            avg_label_val = self.scans[scan_num].data[label_column].mean()
            label_str = "%s =%1.3f" % (label_column, avg_label_val)
            plt.plot(xin, yin + offset * idx, fmt, label=label_str)
            plt.xlabel(x)
            plt.ylabel(y)

        if legend:
            plt.legend()

        if show_plot:
            plt.show(block=False)

        return fh

    def mean_col(self, col):
        r"""Take the mean of a given column in every scan of the collection

        Parameters
        ----------
        col : str
           The name of the column for the mean

        Returns
        -------
        array_like
            an array where each element is the average of the column of a specific
            scan in the collection

        """
        return self.func_col(col, np.mean)

    def func_col(self, col, func):
        r""" apply a function to a column an return the value for each scan

        Parameters
        ----------
        col : str
            The name of the column for the mean

        func: function
            The function to apply

        """
        self.scans_check()
        res = np.empty(self.num_scans)

        for idx, scan_key in enumerate(self.scans.keys()):
            res[idx] = func(self.scans[scan_key].data[col])

        return res

    def min_col(self, col):
        r""" find the minimum of a given column in every scan of the collection

        Parameters
        ----------
        col : str
            The name of the column for the means

        Returns
        -------
        array_like
                an array where each element is the minimum value of the column of a specific
                scan in the collection

        """
        return self.func_col(col, np.min)

    def max_col(self, col):
        r""" find the minimum of a given collum in every scan of the collection

        Parameters
        ----------
        col : str
            The name of the column for the mean

        Returns
        -------
        array_like
                an array where each element is the minimum value of the column of a specific
                scan in the collection

        """
        return self.func_col(col, np.max)

    def pcolor(self, x, y, z='detector', clims=None, color_norm='linear', cmap='jet', show_plot=True):
        r"""Create a false colormap for a coloction of scans.

        The y-direction is always what varies between scans.

        Parameters
        ----------
        x : str
            Name of one of the columns that is in each scan to plot along
            the x-axis.

        y : str
            Name of one of the columns that is in each scan to plot along
            the y-axis. This parameter varies between scans, but is
            constant over an individual scan.

        z : str, optional
            Name of one of the data columns that is in each scan to plot
            along the z-axis. Default: `'detector'`.

        clims : array_like, optional
            An array of two floats, where the first is the minimum color
            scale, the second is the maximum of the color scale.  By
            default the maximum and minimum of all data. Default: None

        color_norm : str, optional
            Select a `'log'` or `'linear'` scale. Default: `'linear'`.

        cmap : str, optional
            A matplotlib colormaps.  Default: `'jet'`.

        """
        self.scans_check()
        fh = plt.figure()

        # calculate y spacing
        meany = self.mean_col(col=y)

        # generate an array for bin boundaries of the y axis
        biny = np.zeros(len(meany) + 1)

        # generate the bin boundaries internal to the array
        biny[1:-1] = (meany[:-1] + meany[1:]) / 2

        # generate the first bin boundary to be the same distance from the mean as the second bin boundary
        biny[0] = 2 * meany[0] - biny[1]

        # generate the last bin boundary to be the same distance from the mean as the next to last.
        biny[-1] = 2 * meany[-1] - biny[-2]

        if clims is None:
            # calculate intensity range
            intens_max = 0.
            intens_min = 1.

            for idx, scan_num in enumerate(self.scans.keys()):
                maxz = self.scans[scan_num].data[z].max()
                minz = self.scans[scan_num].data[z].min()

                if maxz > intens_max:
                    intens_max = maxz

                if (minz < intens_min) & (minz > 0):
                    intens_min = minz
        else:
            intens_max = clims[1]
            intens_min = clims[0]

        for idx, scan_num in enumerate(self.scans.keys()):
            meansx = self.scans[scan_num].data[x]
            zvals = self.scans[scan_num].data[z]
            yvals = np.array([biny[idx], meany[idx], biny[idx + 1]])
            xvals = np.zeros(len(meansx) + 1)
            xvals[1:-1] = (meansx[:-1] + meansx[1:]) / 2.
            xvals[0] = 2 * meansx[0] - xvals[1]
            xvals[-1] = 2 * meansx[-1] - xvals[-2]
            xmat = np.vstack((xvals, xvals, xvals))
            ymat = np.tile(yvals, (len(xvals), 1)).T
            zmat = np.vstack((zvals, zvals))

            if color_norm == 'log':
                plt.pcolor(xmat, ymat, zmat, norm=mpl_c.LogNorm(
                    vmin=intens_min, vmax=intens_max), cmap=cmap)
            else:
                plt.pcolor(xmat, ymat, zmat, vmin=intens_min,
                           vmax=intens_max, cmap=cmap)

        plt.xlabel(x)
        plt.ylabel(y)
        plt.colorbar()

        if show_plot:
            plt.show(block=False)

        return fh
